/* Copyright(C) 2004 Brazil

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA */

#include <stdio.h>
#include <string.h>
#include <unistd.h>
#include <pthread.h>
#include <sys/resource.h>
#include "sym.h"
#include "inv.h"
#include "str.h"
#include "lex.h"
#include "senna_in.h"

/* private classes */

/* b-heap */

typedef struct {
  int n_entries;
  int n_bins;
  sen_inv_cursor **bins;
} cursor_heap;

static inline cursor_heap *
cursor_heap_open(int max)
{
  cursor_heap *h = SEN_MALLOC(sizeof(cursor_heap));
  if (!h) { return NULL; }
  h->bins = SEN_MALLOC(sizeof(sen_inv_cursor *) * max);
  if (!h->bins) { 
    SEN_FREE(h);
    return NULL;
  }
  h->n_entries = 0;
  h->n_bins = max;
  return h;
}

static inline sen_rc
cursor_heap_push(cursor_heap *h, sen_inv *inv, sen_id tid, uint32_t offset)
{
  int n, n2;
  sen_inv_cursor *c, *c2;
  if (h->n_entries >= h->n_bins) {
    int max = h->n_bins * 2;
    sen_inv_cursor **bins = SEN_REALLOC(h->bins, sizeof(sen_inv_cursor *) * max);
    if (!bins) { return sen_memory_exhausted; }
    h->n_bins = max;
    h->bins = bins;
    sen_log("expanded cursor_heap to %d", max);
  }
  if (!(c = sen_inv_cursor_open(inv, tid))) {
    sen_log("cursor open failed");
    return sen_other_error;
  }
  if (sen_inv_cursor_next(c)) {
    sen_log("invalid inv_cursor");
    sen_inv_cursor_close(c);
    return sen_other_error;
  }
  c->offset = offset;
  n = h->n_entries++;
  while (n) {
    n2 = (n - 1) >> 1;
    c2 = h->bins[n2];
    if (c->rid >= c2->rid) { break; }
    h->bins[n] = c2;
    n = n2;
  }
  h->bins[n] = c;
  return sen_success;
}

static inline sen_inv_cursor *
cursor_heap_min(cursor_heap *h)
{
  return h->n_entries ? h->bins[0] : NULL;
}

static inline void
cursor_heap_pop(cursor_heap *h)
{
  int n = 0, n1, n2, m;
  sen_inv_cursor *c = h->bins[0], *c1, *c2;
  if (!h->n_entries) { return; }
  if (sen_inv_cursor_next(c)) {
    sen_inv_cursor_close(c);
    c = h->bins[0] = h->bins[--h->n_entries];
  }
  if ((m = h->n_entries) <= 1) { return; }
  for (;;) {
    n1 = n * 2 + 1;
    n2 = n1 + 1;
    c1 = n1 < m ? h->bins[n1] : NULL;
    c2 = n2 < m ? h->bins[n2] : NULL;
    if (c1 && c->rid > c1->rid) {
      if (c2 && c->rid > c2->rid && c1->rid > c2->rid) {
	h->bins[n] = c2;
	n = n2;
      } else {
	h->bins[n] = c1;
	n = n1;
      }
    } else {
      if (c2 && c->rid > c2->rid) {
	h->bins[n] = c2;
	n = n2;
      } else {
	h->bins[n] = c;
	break;
      }
    }
  }
}

static inline void
cursor_heap_close(cursor_heap *h)
{
  int i = h->n_entries;
  while (i--) { sen_inv_cursor_close(h->bins[i]); }
  SEN_FREE(h->bins);
  SEN_FREE(h);
}

/* b-tree 

struct _cursor_node {
  sen_inv_cursor *c;
  struct _cursor_node *more;
  struct _cursor_node *less;
};

typedef struct _cursor_node cursor_node;

static inline cursor_node *
cursor_node_open(sen_inv *inv, sen_id key) 
{
  cursor_node *n = SEN_MALLOC(sizeof(cursor_node));
  n->c = sen_inv_cursor_open(inv, key);
  if (sen_inv_cursor_next(n->c)) {
    sen_log("invalid inv_cursor");
    sen_inv_cursor_close(n->c);
    SEN_FREE(n);
    return NULL;
  }
  n->more = NULL;
  n->less = NULL;
  return n;
}

static inline void
cursor_node_close(cursor_node *n)
{
  if (!n) { return ; }
  cursor_node_close(n->more);
  cursor_node_close(n->less);
  sen_inv_cursor_close(n->c);
  SEN_FREE(n);
}

static inline void
cursor_node_push(cursor_node **root, cursor_node *node)
{
  cursor_node *r = *root;
  if (!node) { return; }
  if (!r) {
    *root = node;
  } else {
    cursor_node_push(r->c->rid < node->c->rid ? &r->more : &r->less, node);
  }
}

*/

/* token_info */

typedef struct {
  // cursor_node *cursors;
  cursor_heap *cursors;
  int offset;
  uint32_t rid;
  int size;
  int ntoken;
} token_info;

#define EXPAND_NONE 0
#define EXPAND_PREFIX 1
#define EXPAND_SUFFIX 2

#define EXACT_MATCH 0
#define PARTIAL_MATCH 1

inline static void
token_info_expand_both(sen_index *i, const char *key, token_info *ti)
{
  int s = 0;
  sen_set *h, *g;
  uint32_t *offset2;
  sen_set_cursor *c;
  sen_id *tp, *tq;
  if (!(h = sen_sym_prefix_search(i->lexicon, key))) { return; }
  sen_log("key=%s h->n=%d", key, h->n_entries);
  ti->cursors = cursor_heap_open(h->n_entries + 256);
  c = sen_set_cursor_open(h);
  while (sen_set_cursor_next(c, (void **) &tp, NULL)) {
    const char *key2 = sen_sym_key(i->lexicon, *tp);
    sen_log("key2=%s", key2);
    if (sen_str_len(key2, i->lexicon->encoding, NULL) == 1) { // todo: 
      if ((s = sen_inv_estimate_size(i->inv, *tp))) {
	// cursor_node_push(&ti->cursors, cursor_node_open(i->inv, *tp));
	cursor_heap_push(ti->cursors, i->inv, *tp, 0);
	ti->ntoken++;
	ti->size += s;
      }
    } else {
      g = sen_sym_suffix_search(i->lexicon, key2);
      sen_log("g=%d", g ? g->n_entries : 0);
      if (g) {
	sen_set_cursor *d = sen_set_cursor_open(g);
	while (sen_set_cursor_next(d, (void **) &tq, (void **) &offset2)) {
	  if ((s = sen_inv_estimate_size(i->inv, *tq))) {
	    sen_log("est=%d key=%s", s, sen_sym_key(i->lexicon, *tq));
	    // cursor_node_push(&ti->cursors, cursor_node_open(i->inv, *tq));
	    cursor_heap_push(ti->cursors, i->inv, *tq, /* *offset2 */ 0);
	    ti->ntoken++;
	    ti->size += s;
	  }
	}
	sen_set_cursor_close(d);
	sen_set_close(g);
      }
    }
    // if (ti->size > 1000000) { sen_log("huge! %d", ti->size); break; } // todo!!!
  }
  sen_set_cursor_close(c);
  sen_set_close(h);
  sen_log("key=%s done..", key);
}

inline static token_info *
token_info_open(sen_index *i, const char *key, uint32_t offset, int mode)
{
  int s = 0;
  sen_set *h;
  token_info *ti;
  sen_id tid;
  sen_set_cursor *c;
  sen_id *tp;
  if (!(ti = SEN_MALLOC(sizeof(token_info)))) { return NULL; }
  ti->cursors = NULL;
  ti->size = 0;
  ti->ntoken = 0;
  ti->offset = offset;
  switch (mode) {
  case EXPAND_PREFIX|EXPAND_SUFFIX :
    token_info_expand_both(i, key, ti);
    break;
  case EXPAND_NONE :
    if (!(tid = sen_sym_get(i->lexicon, key))) { break; }
    if (!(s = sen_inv_estimate_size(i->inv, tid))) { break; }
    ti->cursors = cursor_heap_open(1);
    cursor_heap_push(ti->cursors, i->inv, tid, 0);
 // cursor_node_push(&ti->cursors, cursor_node_open(i->inv, tid));
    ti->ntoken++;
    ti->size = s;
    break;
  case EXPAND_PREFIX :
    if (!(h = sen_sym_prefix_search(i->lexicon, key))) { break; }
    sen_log("key=%s h->n=%d", key, h->n_entries);
    ti->cursors = cursor_heap_open(h->n_entries);
    c = sen_set_cursor_open(h);
    while (sen_set_cursor_next(c, (void **) &tp, NULL)) {
      if ((s = sen_inv_estimate_size(i->inv, *tp))) {
	sen_log("%8d %s", s, sen_sym_key(i->lexicon, *tp));
	// cursor_node_push(&ti->cursors, cursor_node_open(i->inv, *tp));
	cursor_heap_push(ti->cursors, i->inv, *tp, 0);
	ti->ntoken++;
	ti->size += s;
      }
      //if (ti->size > 1000000) { sen_log("huge! %d", ti->size); break; } // todo!!!
    }
    sen_set_cursor_close(c);
    sen_set_close(h);
    sen_log("key=%s done.", key);
    break;
  case EXPAND_SUFFIX :
    {
      uint32_t *offset2;
      if (!(h = sen_sym_suffix_search(i->lexicon, key))) { break; }
      sen_log("key=%s h->n=%d", key, h->n_entries);
      ti->cursors = cursor_heap_open(h->n_entries);
      c = sen_set_cursor_open(h);
      while (sen_set_cursor_next(c, (void **) &tp, (void **) &offset2)) {
	if ((s = sen_inv_estimate_size(i->inv, *tp))) {
	  // cursor_node_push(&ti->cursors, cursor_node_open(i->inv, *tp));
	  cursor_heap_push(ti->cursors, i->inv, *tp, /* *offset2 */ 0);
	  ti->ntoken++;
	  ti->size += s;
	}
      }
      sen_set_cursor_close(c);
      sen_set_close(h);
      break;
    }
  }
  if (ti->cursors && cursor_heap_min(ti->cursors)) {
    // cursor_node *n;
    // for (n = ti->cursors; n->less; n = n->less);
    // ti->rid = n->c->rid;
    ti->rid = cursor_heap_min(ti->cursors)->rid;
  } else {
    SEN_FREE(ti);
    ti = NULL;
  }
  return ti;
}

static inline sen_rc
token_info_next(token_info *ti)
{
  sen_inv_cursor *c;
  for (;;) {
    if (!(c = cursor_heap_min(ti->cursors))) { return sen_other_error; }
    if (ti->rid != c->rid) { break; }
    cursor_heap_pop(ti->cursors);
  }
  ti->rid = c->rid;
  return sen_success;
}

inline static int
token_info_get_postings(token_info *ti, sen_inv_postings *post, int nterms)
{
  int count = 0;
  /*
  cursor_node *n;
  for (n = ti->cursors; n; n = n->less) {
    if (n->c->rid == ti->rid) {
      count += sen_inv_cursor_get_postings(n->c, post, ti->offset, nterms);
    }
  }
  */
  sen_inv_cursor *c;
  while ((c = cursor_heap_min(ti->cursors))) {
    if (c->rid != ti->rid) { break; }
    count += sen_inv_cursor_get_postings(c, post, ti->offset - c->offset, nterms);
    cursor_heap_pop(ti->cursors);
  }
  return count;
}

inline static int
token_info_init_postings(token_info *ti, sen_inv_postings *post, int nterms)
{
  if (post->mark) {
    //    /* temporally 20041117
    memset(post->array, 0, sizeof(uint32_t) * post->mark);
    //    */
    post->mark = 0;
  }
  return token_info_get_postings(ti, post, nterms);
}

inline static sen_rc
token_info_close(token_info *ti)
{
  //  cursor_node_close(ti->cursors);
  cursor_heap_close(ti->cursors);
  SEN_FREE(ti);
  return sen_success;
}

inline static int
token_compare(const void *a, const void *b)
{
  const token_info *t1 = *((token_info **)a), *t2 = *((token_info **)b);
  return t1->size - t2->size;
}

inline static sen_rc
token_info_build(sen_index *i, const char *query, token_info **tis, uint32_t *n, int flag)
{
  sen_lex *lex;
  token_info *ti;
  sen_id tid;
  sen_rc rc = sen_other_error;
  lex = sen_lex_open(i->lexicon, query, 0);
  tid = sen_lex_next(lex);
  switch (lex->status) {
  case sen_lex_doing :
    ti = token_info_open(i, sen_sym_key(i->lexicon, tid), lex->pos,
		      (flag == EXACT_MATCH) ? EXPAND_NONE : EXPAND_SUFFIX);
    break;
  case sen_lex_done :
    ti = token_info_open(i, sen_sym_key(i->lexicon, tid), lex->pos,
		      (flag == EXACT_MATCH) ? EXPAND_NONE : EXPAND_PREFIX|EXPAND_SUFFIX);
    break;
  case sen_lex_not_found :
    ti = token_info_open(i, lex->str, 0,
		      (flag == EXACT_MATCH) ? EXPAND_NONE : EXPAND_PREFIX|EXPAND_SUFFIX);
    break;
  default :
    goto exit;
  }
  if (!ti) { goto exit ; }
  tis[(*n)++] = ti;
  sen_log("%d:%s", lex->pos, (tid == SEN_SYM_NIL) ? lex->str : sen_sym_key(i->lexicon, tid));
  while (lex->status == sen_lex_doing) {
    tid = sen_lex_next(lex);
    switch (lex->status) {
    case sen_lex_doing :
      ti = token_info_open(i, sen_sym_key(i->lexicon, tid), lex->pos, EXPAND_NONE);
      break;
    case sen_lex_done :
      ti = token_info_open(i, sen_sym_key(i->lexicon, tid), lex->pos,
			   (flag == EXACT_MATCH) ? EXPAND_NONE : EXPAND_PREFIX);
      break;
    default :
      ti = token_info_open(i, lex->next, lex->pos,
			   (flag == EXACT_MATCH) ? EXPAND_NONE : EXPAND_PREFIX);
      break;
    }
    if (!ti) { goto exit; }
    tis[(*n)++] = ti;
    sen_log("%d:%s", lex->pos, (tid == SEN_SYM_NIL) ? lex->next : sen_sym_key(i->lexicon, tid));
  }
  rc = sen_success;
exit :
  sen_lex_close(lex);
  return rc;
}

/* senna.conf */

#define CONF_PATH SENNA_HOME "/senna.conf"

static int sen_partial_match_threshold = 0;
static uint32_t sen_initial_n_segments = 0;
static sen_encoding sen_default_encoding = sen_enc_euc_jp;
static char *initial_n_segments = "INITIAL_N_SEGMENTS";
static char *default_encoding = "DEFAULT_ENCODING";
static char *partial_match_threshold = "PARTIAL_MATCH_THRESHOLD";

#define BUFSIZE 32768

static void 
read_conf(void)
{
  FILE *fp;
  char *buffer, *p;
  if (!(fp = fopen(CONF_PATH, "r"))) { return; }
  if (!(buffer = SEN_MALLOC(BUFSIZE))) { return ; }
  while (!feof(fp)) {
    if (!fgets(buffer, BUFSIZE, fp)) { break; }
    if (!strncmp(buffer, initial_n_segments, strlen(initial_n_segments))) {
      for (p = buffer; *p && (*p < '0' || *p > '9'); p++) {
	if (p - buffer > BUFSIZE) { break; }
      }
      sen_initial_n_segments = atoi(p);
      sen_log("INITIAL_N_SEGMENTS=%d", sen_initial_n_segments);
    }
    if (!strncmp(buffer, partial_match_threshold, strlen(partial_match_threshold))) {
      for (p = buffer; *p && (*p < '0' || *p > '9') && *p != '-'; p++) {
	if (p - buffer > BUFSIZE) { break; }
      }
      sen_partial_match_threshold = atoi(p);
      sen_log("PARTIAL_MATCH_THRESHOLD=%d", sen_partial_match_threshold);
    }
    if (!strncmp(buffer, default_encoding, strlen(default_encoding))) {
      for (p = buffer + strlen(default_encoding); *p; p++) {
	if (p - buffer > BUFSIZE) { break; }
	if (*p == 'e' || *p == 'E') {
	  sen_default_encoding = sen_enc_euc_jp;
	  break;
	}
	if (*p == 'u' || *p == 'U') {
	  sen_default_encoding = sen_enc_utf8;
	  break;
	}
	if (*p == 's' || *p == 'S') {
	  sen_default_encoding = sen_enc_sjis;
	  break;
	}
	if (*p == 'n' || *p == 'N') {
	  sen_default_encoding = sen_enc_none;
	  break;
	}
      }
    }
  }
  SEN_FREE(buffer);
  fclose(fp);
  {
  struct rlimit rlim;
  if (!getrlimit(RLIMIT_STACK, &rlim)) {
    sen_log("RLIMIT_STACK is %d (%d)", rlim.rlim_cur, getpid());
    if (rlim.rlim_cur != RLIM_INFINITY && rlim.rlim_cur < SEN_STACK_SIZE) {
      rlim.rlim_cur = SEN_STACK_SIZE;
      if (!setrlimit(RLIMIT_STACK, &rlim)) {
	sen_log("expanded RLIMIT_STACK to %d", rlim.rlim_cur);
      }
    }
  }
  }
}

/* init */

sen_rc
sen_init(void)
{
  struct rlimit rlim;
  if (!getrlimit(RLIMIT_STACK, &rlim)) {
    sen_log("RLIMIT_STACK is %d (%d)", rlim.rlim_cur, getpid());
    if (rlim.rlim_cur != RLIM_INFINITY && rlim.rlim_cur < SEN_STACK_SIZE) {
      rlim.rlim_cur = SEN_STACK_SIZE;
      if (!setrlimit(RLIMIT_STACK, &rlim)) {
	sen_log("expanded RLIMIT_STACK to %d", rlim.rlim_cur);
      }
    }
  }
  return sen_lex_init();
}

/* sen_index */

sen_index *
sen_index_create(const char *path, int key_size,
		 int flags, int initial_n_segments, sen_encoding encoding)
{
  sen_index *i;
  char buffer[PATH_MAX];
  read_conf();
  if (initial_n_segments == 0) { initial_n_segments = sen_initial_n_segments; }
  if (encoding == sen_enc_default) { encoding = sen_default_encoding; }
  if (strlen(path) > PATH_MAX - 4) {
    sen_log("too long index path (%s)", path);
    return NULL;
  }
  if (!(i = SEN_MALLOC(sizeof(sen_index)))) { return NULL; }
  if (encoding != sen_enc_euc_jp) {
    flags &= ~SEN_INDEX_NORMALIZE;
    /* todo: SEN_INDEX_NORMALIZE is only supported with euc */
  }
  sen_log("creating '%s' encoding=%s initial_n_segments=%d",
	  path, sen_enc_string[encoding], initial_n_segments);
  strcpy(buffer, path);
  strcat(buffer, ".SEN");
  if ((i->keys = sen_sym_create(buffer, key_size, 0, sen_enc_none))) {
    strcpy(buffer, path);
    strcat(buffer, ".SEN.l");
    if ((i->lexicon = sen_sym_create(buffer, 0, SEN_SYM_WITH_SIS|flags, encoding))) {
      strcpy(buffer, path);
      strcat(buffer, ".SEN.i");
      if ((i->inv = sen_inv_create(buffer, 0, initial_n_segments))) {
	// pthread_mutex_init(&i->write_mutex, NULL);
	sen_log("index created (%s) flags=%x", path, i->lexicon->flags);
	return i;
      }
      sen_sym_close(i->lexicon);
    }
    sen_sym_close(i->keys);
  }
  SEN_FREE(i);
  return NULL;
}

sen_index *
sen_index_open(const char *path)
{
  sen_index *i;
  char buffer[PATH_MAX];
  if (strlen(path) > PATH_MAX - 4) {
    sen_log("too long index path (%s)", path);
    return NULL;
  }
  read_conf();
  if (!(i = SEN_MALLOC(sizeof(sen_index)))) { return NULL; }
  strcpy(buffer, path);
  strcat(buffer, ".SEN");
  if ((i->keys = sen_sym_open(buffer))) {
    strcpy(buffer, path);
    strcat(buffer, ".SEN.l");
    if ((i->lexicon = sen_sym_open(buffer))) {
      strcpy(buffer, path);
      strcat(buffer, ".SEN.i");
      if ((i->inv = sen_inv_open(buffer))) {
	// pthread_mutex_init(&i->write_mutex, NULL);
	sen_log("index opened (%s) flags=%x", path, i->lexicon->flags);
	return i;
      }
      sen_sym_close(i->lexicon);
    }
    sen_sym_close(i->keys);
  }
  SEN_FREE(i);
  return NULL;
}

sen_rc
sen_index_close(sen_index *i)
{
  if (!i) { return sen_invalid_argument; }
  // pthread_mutex_destroy(&i->write_mutex);
  sen_sym_close(i->keys);
  sen_sym_close(i->lexicon);
  sen_inv_close(i->inv);
  SEN_FREE(i);
  return sen_success;
}

sen_rc
sen_index_remove(const char *path)
{
  char buffer[PATH_MAX];
  if (!path || strlen(path) > PATH_MAX - 8) { return sen_invalid_argument; }
  snprintf(buffer, PATH_MAX, "%s.SEN", path);
  unlink(buffer);
  snprintf(buffer, PATH_MAX, "%s.SEN.i", path);
  unlink(buffer);
  /* todo: SEN.i.c.0n file may be created */
  snprintf(buffer, PATH_MAX, "%s.SEN.i.c", path);
  unlink(buffer);
  snprintf(buffer, PATH_MAX, "%s.SEN.l", path);
  unlink(buffer);
  return sen_success;
}

sen_rc
sen_index_add(sen_index *i, const void *key, const char *value)
{
  int hint;
  sen_set *h;
  sen_lex *lex;
  sen_set_cursor *c;
  sen_inv_updspec **u;
  sen_id rid, tid, *tp;

  // pthread_mutex_lock(&i->write_mutex);

  rid = sen_sym_add(i->keys, key);
  h = sen_set_open(sizeof(sen_id), sizeof(sen_inv_updspec *), 0);
  lex = sen_lex_open(i->lexicon, value, 1);
  while (!lex->status) {
    if ((tid = sen_lex_next(lex))) {
      sen_set_add(h, &tid, (void **) &u);
      if (!*u) { *u = sen_inv_updspec_open(0, rid, 1); }
      // sen_log("(%s) %d", sen_sym_key(i->lexicon, tid), sen_lex_boundary_info(lex));
      sen_inv_updspec_add(*u, lex->pos, 0, sen_lex_boundary_info(lex));
    }
  }
  sen_lex_close(lex);
  c = sen_set_cursor_open(h);
  while (sen_set_cursor_next(c, (void **)&tp, (void **) &u)) {
    if (*tp < 1) {
      // sen_log("%d '%s'", *tp, sen_sym_key(i->lexicon, *tp));
    } else {
      hint = sen_str_get_prefix_order(sen_sym_key(i->lexicon, *tp));
      if (hint == -1) { hint = *tp; }
      sen_inv_update(i->inv, *tp, *u, hint);
      // sen_log("(%s)", sen_sym_key(i->lexicon, *tp));
    }
    sen_inv_updspec_close(*u);
  }

  // pthread_mutex_unlock(&i->write_mutex);
  sen_set_cursor_close(c);
  sen_set_close(h);
  return sen_success;
}

sen_rc
sen_index_del(sen_index *i, const void *key, const char *value)
{
  sen_set *h;
  sen_lex *lex;
  sen_set_cursor *c;
  sen_inv_updspec **u;
  sen_id rid, tid, *tp;

  // sen_log("deleting(%s)", value);
  if (!(rid = sen_sym_get(i->keys, key))) { return sen_invalid_argument; }
  h = sen_set_open(sizeof(sen_id), sizeof(sen_inv_updspec *), 0);
  lex = sen_lex_open(i->lexicon, value, 0);
  while (!lex->status) {
    if ((tid = sen_lex_next(lex))) {
      sen_set_add(h, &tid, (void **) &u);
      if (!*u) { *u = sen_inv_updspec_open(0, rid, 0); }
    }
  }
  sen_lex_close(lex);
  c = sen_set_cursor_open(h);
  // pthread_mutex_lock(&i->write_mutex);
  while (sen_set_cursor_next(c, (void **)&tp, (void **) &u)) {
    if (*tp < 1) {
      sen_log("%d '%s'", *tp, sen_sym_key(i->lexicon, *tp));
    } else {
      sen_inv_delete(i->inv, *tp, *u);
    }
    sen_inv_updspec_close(*u);
  }
  // pthread_mutex_unlock(&i->write_mutex);
  sen_set_cursor_close(c);
  sen_set_close(h);
  return sen_success;
}

sen_rc 
sen_index_upd(sen_index *i, const void *key,
		  const char *oldvalue, const char *newvalue)
{
  // if (!strcmp(oldvalue, newvalue)) { return sen_success; }
  sen_rc rc = sen_success;
  if (oldvalue) { rc = sen_index_del(i, key, oldvalue); }
  if (rc != sen_success) { return rc; }
  if (newvalue) { rc = sen_index_add(i, key, newvalue); }
  return rc;
}

#define MAX_DOCUMENT_SIZE 1048576

void
_sen_index_sel(sen_index *i, const char *query, sen_set *res,
	       token_info **tis, sen_inv_postings *p, int flag)
{
  token_info *ti, **tip;
  uint32_t n = 0, j = 0, rid, nrid, noccur, *score;
  if (token_info_build(i, query, tis, &n, flag)) { goto exit; }
  sen_log("n=%d", n);
  if (!n) { goto exit; }
  qsort(tis, n, sizeof(token_info *), token_compare);
  for (tip = tis, j = 0; j < n; j++, tip++) {
    ti = *tip;
    sen_log("o=%d n=%d s=%d r=%d", ti->offset, ti->ntoken, ti->size, ti->rid);
  }
  for (;;) {
    rid = (*tis)->rid;
    noccur = token_info_init_postings(*tis, p, n);
    for (tip = tis + 1, j = 1, nrid = rid + 1; j < n; j++, tip++) {
      ti = *tip;
      while (ti->rid < rid) {
        if (token_info_next(ti)) { goto exit; }
      }
      if (ti->rid > rid) {
	nrid = ti->rid;
	break;
      }
      noccur = token_info_get_postings(ti, p, n);
    }
    if (noccur) {
      sen_set_add(res, &rid, (void **) &score);
      *score = noccur;
    }
    while ((*tis)->rid < nrid) {
      if (token_info_next(*tis)) { goto exit; }
    }
  }
exit :
  for (tip = tis, j = 0; j < n; j++, tip++) {
    if (*tip) { token_info_close(*tip); }
  }
}

sen_records *
sen_index_sel(sen_index *i, const char *query)
{
  token_info **tis;
  sen_inv_postings p;
  sen_set *res = NULL;
  p.array = NULL;
  sen_log("in (%s)", query);
  if (!(tis = SEN_MALLOC(sizeof(token_info *) * strlen(query) * 2))) { return NULL; }
  p.array = SEN_MALLOC(sizeof(uint32_t) * MAX_DOCUMENT_SIZE);
  p.mark = MAX_DOCUMENT_SIZE;
  p.size = MAX_DOCUMENT_SIZE;
  res = sen_set_open(sizeof(int), sizeof(int), 0);

  _sen_index_sel(i, query, res, tis, &p, EXACT_MATCH);
  sen_log("exact: %d", res->n_entries);

  if (res->n_entries <= sen_partial_match_threshold) {
    _sen_index_sel(i, query, res, tis, &p, PARTIAL_MATCH);
    sen_log("partial: %d", res->n_entries);
  }

  SEN_FREE(tis); 
  if (p.array) { SEN_FREE(p.array); }
  sen_log("hits=%d", res ? res->n_entries : 0);
  if (res) {
    if (res->n_entries) {
      sen_records *r = SEN_MALLOC(sizeof(sen_records));
      r->index = i;
      r->records = res;
      r->cursor = NULL;
      r->curr_key = NULL;
      r->curr_score = 0;
      return r;
    } else {
      sen_set_close(res);
    }
  }
  return NULL;
}

sen_rc
sen_records_rewind(sen_records *r)
{
  if (!r) { return sen_invalid_argument; }
  if (r->cursor) { sen_set_cursor_close(r->cursor); }
  r->cursor = sen_set_cursor_open(r->records);
  r->curr_key = NULL;
  r->curr_score = 0;
  return sen_success;
}

const void *
sen_records_next(sen_records *r)
{
  sen_id *rid;
  uint32_t *score;
  if (!r) { return NULL; }
  if (!r->cursor) { sen_records_rewind(r); }
  if (sen_set_cursor_next(r->cursor, (void **) &rid, (void **) &score)) {
    r->curr_score = *score;
    r->curr_key = sen_sym_key(r->index->keys, *rid);
    return r->curr_key;
  }
  return NULL;
}

int
sen_records_find(sen_records *r, const void *key)
{
  sen_id rid;
  uint32_t *score;
  if (!r || !r->index || !r->index->keys) { return 0; }
  if (!(rid = sen_sym_get(r->index->keys, key))) { return 0; }
  if (!sen_set_get(r->records, &rid, (void **) &score)) { return 0; }
  return *score;
}

int
sen_records_nhits(sen_records *r)
{
  return r->records->n_entries;
}

int
sen_records_curr_score(sen_records *r)
{
  return r->curr_score;
}

const void *
sen_records_curr_key(sen_records *r)
{
  return r->curr_key;
}

sen_rc
sen_records_close(sen_records *r)
{
  if (!r) { return sen_invalid_argument; }
  if (r->cursor) { sen_set_cursor_close(r->cursor); }
  sen_set_close(r->records);
  SEN_FREE(r);
  return sen_success;
}
