/* Copyright(C) 2004 Brazil

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA */

#include <fcntl.h>
#include <unistd.h>
#include <string.h>
#include <sys/stat.h>
#include <sys/param.h>

#include "str.h"
#include "inv.h"

#define SEN_INV_IDSTR "SENNA:INV:00.00"
#define SEN_INV_HEADER_SIZE  0x10000
#define SEN_INV_SEGMENT_SIZE 0x40000
#define SEN_INV_CHUNK_SIZE   0x40000
#define FILESIZE_MAX 0x80000000
#define MAX_SEGMENT (FILESIZE_MAX / SEN_INV_SEGMENT_SIZE)
#define MAX_ID (FILESIZE_MAX / sizeof(uintptr_t) - 1)
#define W_OF_SEGMENT 18
#define W_OF_ARRAY (W_OF_SEGMENT - 2)
#define ARRAY_MASK_IN_A_SEGMENT (SEN_INV_SEGMENT_SIZE / sizeof(uintptr_t) - 1)
#define BUFFER_MASK_IN_A_SEGMENT (SEN_INV_SEGMENT_SIZE - 1)
#define NCHUNK_IN_A_FILE (FILESIZE_MAX / SEN_INV_CHUNK_SIZE)
#define CHUNK_NOT_ASSIGNED MAX_SEGMENT
#define SEGMENT_ARRAY 0x8000
#define SEGMENT_BUFFER 0x4000
#define SEGMENT_MASK (MAX_SEGMENT - 1)

#define B30 0x40000000
#define B31 0x80000000
#define BIT10_01(x) ((x >> 1) & 0x3ff)
#define BIT17_00(x) (x & 0x3ffff)
#define BIT19_18(x) ((x >> 18) & 0x03)
#define BIT26_00(x) (x & 0x7ffffff)
#define BIT28_27(x) ((x >> 27) & 0x03)
#define BIT29_11(x) ((x >> 11) & 0x7ffff)
#define BIT29_27(x) ((x >> 27) & 0x07)
#define BIT30_18(x) ((x >> 18) & 0x1fff)
#define BIT30_20(x) ((x >> 20) & 0x07ff)
#define BIT30_27(x) ((x >> 27) & 0x0f)
#define BIT30_29(x) ((x >> 29) & 0x03)
#define BIT31_18(x) (x >> 18)
#define BIT31_29(x) (x >> 29)
#define BIT31_30(x) (x >> 30)
#define BIT30(x) (x & B30)
#define BIT31(x) (x & B31)

#define SEN_INV_INITIAL_N_SEGMENTS 512

#define NEXT_ADDR(p) (((void *)p) + sizeof *p)

struct sen_inv_header {
  char idstr[16];
  uint16_t segments[MAX_SEGMENT];
  uint8_t chunks[MAX_SEGMENT];
  uint32_t initial_n_segments;
};

struct _sen_inv_pos {
  struct _sen_inv_pos *next;
  uint32_t pos;
};

struct _sen_inv_updspec {
  uint8_t type;
  int_least8_t boundary;
  uint32_t rid;
  int32_t score;
  int32_t tf;
  struct _sen_inv_pos *pos;
  struct _sen_inv_pos *tail;
};

struct _sen_inv {
  sen_io *seg;
  sen_io *chunk;
  struct sen_inv_header *header;
  sen_io_seginfo ainfo[MAX_SEGMENT];
  sen_io_seginfo binfo[MAX_SEGMENT];
};

struct _sen_inv_entry {
  /* iroiro.. */
  int rc;
};

/* encode/decode */

#define ENCODE(v,p) \
{ \
  uint8_t *_p = (uint8_t *)p; \
  uint32_t _v = v; \
  if (_v < 0x8f) { \
    *_p++ = _v; \
  } else if (_v < 0x408f) { \
    _v -= 0x8f; \
    *_p++ = 0xc0 + (_v >> 8); \
    *_p++ = _v & 0xff; \
  } else if (_v < 0x20408f) { \
    _v -= 0x408f; \
    *_p++ = 0xa0 + (_v >> 16); \
    *_p++ = (_v >> 8) & 0xff; \
    *_p++ = _v & 0xff; \
  } else if (_v < 0x1020408f) { \
    _v -= 0x20408f; \
    *_p++ = 0x90 + (_v >> 24); \
    *_p++ = (_v >> 16) & 0xff; \
    *_p++ = (_v >> 8) & 0xff; \
    *_p++ = _v & 0xff; \
  } else { \
    *_p++ = 0x8f; \
    *((uint32_t *)_p) = _v; \
    _p += sizeof(uint32_t); \
  } \
  p = _p; \
}

#define DECODE(v,p) \
{ \
  uint8_t *_p = (uint8_t *)p; \
  uint32_t _v = *_p++; \
  switch (_v >> 4) { \
  case 0x08 : \
    if (_v == 0x8f) { \
      _v = *((uint32_t *)_p); \
      _p += sizeof(uint32_t); \
    } \
    break; \
  case 0x09 : \
    _v = (_v - 0x90) * 0x100 + *_p++; \
    _v = _v * 0x100 + *_p++; \
    _v = _v * 0x100 + *_p++ + 0x20408f; \
    break; \
  case 0x0a : \
  case 0x0b : \
    _v = (_v - 0xa0) * 0x100 + *_p++; \
    _v = _v * 0x100 + *_p++ + 0x408f; \
    break; \
  case 0x0c : \
  case 0x0d : \
  case 0x0e : \
  case 0x0f : \
    _v = (_v - 0xc0) * 0x100 + *_p++ + 0x8f; \
    break; \
  } \
  v = _v; \
  p = _p; \
}

#define SKIP(p) \
{ \
  uint8_t *_p = (uint8_t *)p; \
  uint32_t _v = *_p++; \
  switch (_v >> 4) { \
  case 0x08 : \
    if (_v == 0x8f) { \
      _p += sizeof(uint32_t); \
    } \
    break; \
  case 0x09 : \
    _p += 3; \
    break; \
  case 0x0a : \
  case 0x0b : \
    _p += 2; \
    break; \
  case 0x0c : \
  case 0x0d : \
  case 0x0e : \
  case 0x0f : \
    _p += 1; \
    break; \
  } \
  p = _p; \
}

/* segment */

inline static sen_rc
segment_get(sen_inv *inv, uint16_t type, uint32_t segno, sen_io_seginfo *si)
{
  uint16_t s;
  int i, empty = MAX_SEGMENT;
  for (i = 0; i < MAX_SEGMENT; i++) {
    s = inv->header->segments[i];
    if (s) {
      if (s == (type | segno)) { break; }
    } else {
      if (empty == MAX_SEGMENT) { empty = i; }
    }
  }
  if (i == MAX_SEGMENT) {
    if (empty == MAX_SEGMENT) { return sen_memory_exhausted; }
    inv->header->segments[empty] = type | segno;
    si->segno = empty;
    if (sen_io_seg_bind(inv->seg, si)) {
      return sen_memory_exhausted;
    }
    memset(si->addr, 0, SEN_INV_SEGMENT_SIZE);
  } else {
    si->segno = i;
  }
  return sen_success;
}

inline static sen_rc
segment_new(sen_inv *inv, uint16_t type, uint32_t *segno)
{
  uint16_t s;
  int i, seg, empty = MAX_SEGMENT;
  sen_rc rc = sen_success;
  char used[MAX_SEGMENT];
  memset(used, 0, MAX_SEGMENT);
  for (i = 0; i < MAX_SEGMENT; i++) {
    if ((s = inv->header->segments[i])) {
      if (s & type) { used[s & SEGMENT_MASK]++; }
    } else {
      if (empty == MAX_SEGMENT) { empty = i; }
    }
  }
  if (empty == MAX_SEGMENT) { return sen_memory_exhausted; }
  if (segno && *segno < MAX_SEGMENT) {
    if (used[*segno]) { return sen_invalid_argument; }
    seg = *segno;
  } else {
    for (seg = 0; used[seg]; seg++) ;
  }
  inv->header->segments[empty] = type | seg;
  switch (type) {
  case SEGMENT_ARRAY :
    inv->ainfo[seg].segno = empty;
    break;
  case SEGMENT_BUFFER :
    inv->binfo[seg].segno = empty;
    break;
  }
  if (segno) { *segno = seg; }
  return rc;
}

inline static sen_rc
load_all_segments(sen_inv *inv)
{
  sen_rc rc = sen_success;
  int seg;
  uint16_t s;
  for (seg = 0; seg < MAX_SEGMENT; seg++) {
    if (!(s = inv->header->segments[seg])) { continue; }
    if (s & SEGMENT_ARRAY) {
      inv->ainfo[s & SEGMENT_MASK].segno = seg;
    }
    if (s & SEGMENT_BUFFER) {
      inv->binfo[s & SEGMENT_MASK].segno = seg;
    }
  }
  return rc;
}

/* chunk */

inline static sen_rc
chunk_new(sen_inv *inv, uint32_t *res, uint32_t size)
{
  int i, j, n = size / SEN_INV_CHUNK_SIZE;
  if (n * SEN_INV_CHUNK_SIZE < size) { n++; }
  for (i = 0, j = -1; i < MAX_SEGMENT; i++) {
    if (inv->header->chunks[i]) {
      j = i;
    } else {
      if (i - j == n) {
	if (res) { *res = j + 1; }
	while (j < i) {
	  inv->header->chunks[++j] = 1;
	}
	return sen_success;
      }
    }
  }
  return sen_memory_exhausted;
}

inline static sen_rc
chunk_free(sen_inv *inv, int start, uint32_t size)
{
  int i, n = size / SEN_INV_CHUNK_SIZE;
  if (n * SEN_INV_CHUNK_SIZE < size) { n++; }

  sen_log("chunk_free start=%d size=%d(%d)", start, size, n);

  for (i = 0; i < n; i++) {
    inv->header->chunks[start + i] = 0;
  }
  return sen_success;
}

/* buffer */

typedef struct {
  uint32_t tid;
  uint32_t size_in_chunk;
  uint32_t pos_in_chunk;
  uint32_t pos_in_buffer;
} buffer_term;

typedef struct {
  uint32_t var0;
  uint32_t var1;
} buffer_rec;

typedef struct {
  uint32_t chunk;
  uint32_t chunk_size;
  uint32_t buffer_free;
  uint32_t nterms;
} buffer_header;

typedef struct {
  buffer_header header;
  buffer_term terms[(SEN_INV_SEGMENT_SIZE - sizeof(buffer_header))/sizeof(buffer_term)];
} buffer;

inline static sen_rc
buffer_at(sen_inv *inv, uint32_t pos, buffer_term **bt, buffer **b)
{
  sen_io_seginfo *si = &inv->binfo[pos >> W_OF_SEGMENT];
  if (!si->addr) {
    if (si->segno == -1) { load_all_segments(inv); }
    if (sen_io_seg_bind(inv->seg, si)) { return sen_memory_exhausted; }
  }
  if (b) { *b = si->addr; }
  if (bt) { *bt = (buffer_term *)(si->addr + (pos & BUFFER_MASK_IN_A_SEGMENT)); }
  return sen_success;
}

inline static sen_rc
buffer_alloc(sen_inv *inv, uint32_t size, uint32_t *pos,
	     buffer_term **bt, buffer_rec **br, buffer **bp, int hint)
{
  int seg;
  buffer *b;
  if (size + sizeof(buffer_header) + sizeof(buffer_term) > SEN_INV_SEGMENT_SIZE) {
    return sen_invalid_argument;
  }
  seg = hint % inv->header->initial_n_segments;
  if ((b = inv->binfo[seg].addr)) {
    if (b->header.buffer_free >= size + sizeof(buffer_term)) { goto exit; }
    while (++seg < MAX_SEGMENT) {
      if (!(b = inv->binfo[seg].addr)) { continue; } // todo: only from cache?
      if (b->header.chunk == CHUNK_NOT_ASSIGNED // todo: more intelligent
	  && b->header.nterms < 1024
	  && b->header.buffer_free >= size + sizeof(buffer_term)) {
	goto exit;
      }
    }
  }
  if (segment_new(inv, SEGMENT_BUFFER, &seg) || 
      buffer_at(inv, seg * SEN_INV_SEGMENT_SIZE, NULL, &b)) {
    return sen_memory_exhausted;
  }
  memset(b, 0, SEN_INV_SEGMENT_SIZE);
  b->header.buffer_free = SEN_INV_SEGMENT_SIZE - sizeof(buffer_header);
  b->header.chunk = CHUNK_NOT_ASSIGNED;
  b->header.chunk_size = 0;
exit :  
  if (pos) {
    *pos = seg * SEN_INV_SEGMENT_SIZE
      + sizeof(buffer_header)
      + sizeof(buffer_term) * b->header.nterms;
  }
  if (bt) { *bt = &b->terms[b->header.nterms];}
  b->header.nterms++;
  b->header.buffer_free -= size + sizeof(buffer_term);
  if (br) { *br = ((void *)&b->terms[b->header.nterms]) + b->header.buffer_free; }
  if (bp) { *bp = b; }
  return sen_success;
}

/* array */

inline static uint32_t *
array_at(sen_inv *inv, uint32_t id)
{
  sen_io_seginfo *si;
  if (id > MAX_ID) { return NULL; }
  si = &inv->ainfo[id >> W_OF_ARRAY];
  if (!si->addr) {
    if (si->segno == -1) { load_all_segments(inv); }
    if (sen_io_seg_bind(inv->seg, si)) { return NULL; }
  }
  return (uint32_t*)(si->addr + (id & ARRAY_MASK_IN_A_SEGMENT) * sizeof(uint32_t));
}

inline static uint32_t *
array_get(sen_inv *inv, uint32_t id)
{
  sen_io_seginfo *si;
  uint32_t seg;
  if (id > MAX_ID) { return NULL; }
  seg = id >> W_OF_ARRAY;
  si = &inv->ainfo[seg];
  if (!si->addr) {
    if (si->segno == -1) {
      if (segment_get(inv, SEGMENT_ARRAY, seg, si)) { return NULL; }
    }
    if (sen_io_seg_bind(inv->seg, si)) { return NULL; }
  }
  return (uint32_t*)(si->addr + (id & ARRAY_MASK_IN_A_SEGMENT) * sizeof(uint32_t));
}

/* updspec */

sen_inv_updspec *
sen_inv_updspec_open(uint8_t type, uint32_t rid, int32_t score)
{
  sen_inv_updspec *u;
  if (!(u = SEN_MALLOC(sizeof(sen_inv_updspec)))) { return NULL; }
  u->type = type;
  u->rid = rid;
  u->boundary = 0x03;
  u->score = score;
  u->tf = 0;
  u->pos = NULL;
  u->tail = NULL;
  return u;
}

sen_rc
sen_inv_updspec_close(sen_inv_updspec *u)
{
  struct _sen_inv_pos *p = u->pos, *q;
  while (p) {
    q = p->next;
    SEN_FREE(p);
    p = q;
  }
  SEN_FREE(u);
  return sen_success;
}

inline static buffer_rec *
encode_rec(sen_inv_updspec *u, int *size)
{
  uint8_t *p;
  uint32_t lpos;
  buffer_rec *br;
  struct _sen_inv_pos *pp;
  if (!(br = SEN_MALLOC(sizeof(buffer_rec) + 5 * (u->tf + 2)))) {
    return NULL;
  }
  if (u->tf == 1 && u->score < 4 && u->pos->pos < 0x4000) {
    /*
    struct buffer_rec0 {
      union {
        uint32_t var0;
	struct {
	  unsigned int flag0 : 1 = 1;
	  unsigned int score : 2;
	  unsigned int boundary : 2;
	  unsigned int rid : 27;
        };
      };
      union {
        uint32_t var1;
	struct {
	  unsigned int pos : 14;
	  uint32_t next : 18;
        };
      };
    };
    */
    br->var0 = B31 + (u->score << 29) + (u->boundary << 27) + u->rid;
    br->var1 = u->pos->pos << 18;
    *size = sizeof(buffer_rec);
  } else if (u->tf == 2 && u->score < 4 && u->pos->pos < 0x2000) {
    /*
    struct buffer_rec0 {
      union {
        uint32_t var0;
	struct {
	  unsigned int flag0 : 1 = 0;
	  unsigned int score : 2;
	  unsigned int boundary : 2;
	  unsigned int rid : 27;
        };
      };
      union {
        uint32_t var1;
	struct {
	  unsigned int flag1 : 1 = 1;
	  unsigned int pos : 13;
	  uint32_t next : 18;
        };
      };
    };
    */
    br->var0 = (u->score << 29) + (u->boundary << 27) + u->rid;
    br->var1 = B31 + (u->pos->pos << 18);
    p = NEXT_ADDR(br);
    ENCODE(u->pos->next->pos - u->pos->pos, p);
    *size = (void *)p - (void *)br;
  } else if (u->tf < 0x08 && u->score < 0x800) {
    /*
    struct buffer_rec0 {
      union {
        uint32_t var0;
	struct {
	  unsigned int flag0 : 2 = 01;
	  unsigned int tf : 3;
	  unsigned int rid : 27;
        };
      };
      union {
        uint32_t var1;
	struct {
	  unsigned int flag1 : 1 = 0;
	  unsigned int score : 11;
	  unsigned int boundary : 2;
	  uint32_t next : 18;
        };
      };
    };
    */
    br->var0 = B30 + (u->tf << 27) + u->rid;
    br->var1 = (u->score << 20) + (u->boundary << 18);
    p = NEXT_ADDR(br);
    for (lpos = 0, pp = u->pos; pp; lpos = pp->pos, pp = pp->next) {
      ENCODE(pp->pos - lpos, p);
    }
    *size = (void *)p - (void *)br;
  } else {
    /*
    struct buffer_rec0 {
      union {
        uint32_t var0;
	struct {
          unsigned int flag0 : 2 = 00;
	  unsigned int _brank : 3;
          unsigned int rid : 27;
        };
      };
      union {
        uint32_t var1;
	struct {
          unsigned int flag1 : 1 = 0;
	  unsigned int _brank : 13;
          uint32_t next : 18;
        };
      };
    };
    */
    br->var0 = u->rid;
    br->var1 = 0;
    p = NEXT_ADDR(br);
    ENCODE(u->tf, p);
    ENCODE((u->score << 2) + u->boundary, p);
    for (lpos = 0, pp = u->pos; pp; lpos = pp->pos, pp = pp->next) {
      ENCODE(pp->pos - lpos, p);
    }
    *size = (void *)p - (void *)br;
  }
  return br;
}

struct _sen_inv_set_entry {
  uint32_t rid;
  buffer_rec *br;
};

inline static sen_rc
buffer_flush(sen_inv *inv, uint32_t seg)
{
  sen_io_seginfo *si;
  buffer *sb, *db;
  sen_io_win sw, dw;
  uint8_t *dc, *sc = NULL;
  uint32_t scn, dcn, i, ss, ds, max_dest_chunk_size;
  si = &inv->binfo[seg];
  if ((ss = si->segno) == -1) { return sen_invalid_format; }
  if (buffer_at(inv, seg * SEN_INV_SEGMENT_SIZE, NULL, &sb)) {
    return sen_memory_exhausted;
  }
  for (ds = 0; inv->header->segments[ds];) {
    if (++ds == MAX_SEGMENT) { return sen_memory_exhausted; }
  }
  if (!(db = sen_io_seg_map(inv->seg, ds))) { return sen_memory_exhausted; }
  memset(db, 0, SEN_INV_SEGMENT_SIZE);

  max_dest_chunk_size = sb->header.chunk_size + SEN_INV_SEGMENT_SIZE;
  if (chunk_new(inv, &dcn, max_dest_chunk_size)) {
    int i, j;
    sen_log("chunk_new failed %d(%d)", max_dest_chunk_size, max_dest_chunk_size / SEN_INV_SEGMENT_SIZE);
    for (i = 0; i < 128; i++) {
      for (j = 0; j < 64; j++) {
        sen_log("%d", inv->header->chunks[i * 64 + j]);
      }
      sen_log(":");
    }
    exit(1);
    return sen_memory_exhausted;
  }
  //  sen_log("db=%p ds=%d sb=%p seg=%d", db, ds, sb, seg);
  if ((scn = sb->header.chunk) != CHUNK_NOT_ASSIGNED) {
    sc = sen_io_win_map(inv->chunk, &sw, scn, 0, sb->header.chunk_size, sen_io_rdwr);
    if (!sc) {
      sen_log("io_win_map(%d, %d) failed!!", scn, sb->header.chunk_size);
      return sen_memory_exhausted;
    }
  }
  // dc = sen_io_win_map(inv->chunk, &dw, dcn, 0, max_dest_chunk_size, sen_io_wronly);
  dc = sen_io_win_map(inv->chunk, &dw, dcn, 0, max_dest_chunk_size, sen_io_rdwr);
  if (!dc) { 
    sen_log("io_win_map(%d, %d) failed!!", dcn, max_dest_chunk_size);
    return sen_memory_exhausted;
  }
  {
    sen_set *h;
    uint8_t *sp = NULL, *spe = NULL, *dp = dc;
    struct _sen_inv_set_entry **hh, *hp = NULL;
    uint32_t next, rid, bdf;
    buffer_rec *br, **brp;
    buffer_term *bt;
    int n = sb->header.nterms;
    memcpy(db->terms, sb->terms, n * sizeof(buffer_term));
    // sen_log(" scn=%d, dcn=%d, nterms=%d", sb->header.chunk, dcn, n);
    for (bt = db->terms; n; n--, bt++) {
      uint32_t ib = 0, ndf = 0, srid = 0, lrid = 0, brid, tf, score_bound, gap;
      uint8_t *p;
      h = sen_set_open(sizeof(int), sizeof(buffer_rec *), 0);
      for (next = bt->pos_in_buffer; next; next = BIT17_00(br->var1)) {
	br = (buffer_rec *)(((void *)sb) + next);
	rid = BIT26_00(br->var0);
	if (!sen_set_get(h, &rid, NULL)) {
	  sen_set_add(h, &rid, (void **)&brp);
	  *brp = br;
	}
      }
      if (sc) { 
	sp = sc + bt->pos_in_chunk;
	spe = sp + bt->size_in_chunk;
      }
      bdf = h->n_entries;
      bt->pos_in_chunk = dp - dc;
      bt->pos_in_buffer = 0;
      hh = (struct _sen_inv_set_entry **)sen_set_sort(h, 0, NULL);

      // sen_log("db=%p n=%d, bt=%p tid=%d bdf=%d", db, n, bt, bt->tid, bdf);

#define GETNEXTS() { \
  if (sp < spe) { \
    if (srid) { \
      DECODE(tf, sp); \
      DECODE(score_bound, sp); \
      while (tf--) { DECODE(gap, sp); } \
    } \
    DECODE(gap, sp); \
    srid += gap; \
  } else { \
    srid = 0; \
  } \
}
#define GETNEXTB() { \
  if (ib < bdf) { \
    hp = hh[ib++]; \
    brid = hp->rid; \
  } else { \
    brid = 0; \
  } \
}
#define PUTNEXTS() { \
  if (srid) { \
    /* sen_log("srid=%d", srid); */ \
    gap = srid - lrid; \
    ndf++; \
    ENCODE(gap, dp); \
    DECODE(tf, sp); \
    ENCODE(tf, dp); \
    DECODE(score_bound, sp); \
    ENCODE(score_bound, dp); \
    while (tf--) { DECODE(gap, sp); ENCODE(gap, dp);\
    } \
    lrid = srid; \
  } \
  if (sp < spe) { \
    DECODE(gap, sp); \
    srid += gap; \
  } else { \
    srid = 0; \
  } \
  /* sen_log("gap=%d srid=%d", gap, srid); */ \
}
#define PUTNEXTB() { \
  br = hp->br; \
  /* skip 0 score entry */ \
  if (BIT31_29(br->var0) != 0x04) { \
    /* sen_log("brid=%d", brid);  */ \
    gap = brid - lrid; \
    if (gap <= 0) { sen_log("(%d) <= (%d)", hp->rid, lrid); } \
    ndf++; \
    ENCODE(gap, dp); \
    if (BIT31(br->var0)) { \
      ENCODE(1, dp); \
      ENCODE(BIT30_27(br->var0), dp); \
      ENCODE(BIT31_18(br->var1), dp); \
    } else if (BIT31(br->var1)) { \
      ENCODE(2, dp); \
      ENCODE(BIT30_27(br->var0), dp); \
      ENCODE(BIT30_18(br->var1), dp); \
      p = NEXT_ADDR(br); \
      DECODE(gap, p); \
      ENCODE(gap, dp); \
    } else if (BIT30(br->var0)) { \
      tf = BIT29_27(br->var0); \
      ENCODE(tf, dp); \
      ENCODE(BIT31_18(br->var1), dp); \
      p = NEXT_ADDR(br); \
      for (i = 0; i < tf; i++) { \
	DECODE(gap, p); \
	ENCODE(gap, dp); \
      } \
    } else { \
      p = NEXT_ADDR(br); \
      DECODE(tf, p); \
      DECODE(score_bound, p); \
      ENCODE(tf, dp); \
      ENCODE(score_bound, dp); \
      for (i = 0; i < tf; i++) { \
	DECODE(gap, p); \
	ENCODE(gap, dp); \
      } \
    } \
    lrid = brid; \
  } \
  if (ib < bdf) { \
    hp = hh[ib++]; \
    brid = hp->rid; \
  } else { \
    brid = 0; \
  } \
}
      GETNEXTS();
      GETNEXTB();
      for (;;) {
	if (brid) {
	  if (srid) {
	    if (brid > srid) {
	      PUTNEXTS();
	    } else {
	      if (brid == srid) { GETNEXTS(); }
	      PUTNEXTB();
	    }
	  } else {
	    PUTNEXTB();
	  }
	} else {
	  if (srid) {
	    PUTNEXTS();
	  } else {
	    break;
	  }
	}
      }
      // sen_log("break: dp=%p sp=%p", dp, sp);
      if (hh) { SEN_FREE(hh); }
      
      sen_set_close(h);

      bt->size_in_chunk = (dp - dc) - bt->pos_in_chunk;

      // sen_log("db=%p df=%d size=%d", db, ndf, (dp - dc) - bt->pos_in_chunk);

    }
    db->header.chunk_size = dp - dc;
  }
  db->header.chunk = dcn;
  db->header.buffer_free = SEN_INV_SEGMENT_SIZE
    - sizeof(buffer_header) - sb->header.nterms * sizeof(buffer_term);
  db->header.nterms = sb->header.nterms;
  sen_io_seg_unbind(inv->seg, si);
  si->segno = ds;
  {
    int mc, ec;
    mc = max_dest_chunk_size / SEN_INV_CHUNK_SIZE;
    if (mc * SEN_INV_CHUNK_SIZE < max_dest_chunk_size) { mc++; }
    ec = db->header.chunk_size / SEN_INV_CHUNK_SIZE;
    if (ec * SEN_INV_CHUNK_SIZE < db->header.chunk_size) { ec++; }
    // sen_log(" ss=%d ds=%d inv->binfo[%d]=%p max_size=%d(%d) chunk_size=%d(%d)", ss, ds, seg, db, max_dest_chunk_size, mc, db->header.chunk_size, ec);
    while (ec < mc) {
      sen_log("chunk[%d]=0(%d)", ec, mc);
      inv->header->chunks[db->header.chunk + ec++] = 0;
    }
  }
  inv->header->segments[ss] = 0;
  inv->header->segments[ds] = SEGMENT_BUFFER | seg;

  if (scn != CHUNK_NOT_ASSIGNED) {
    sen_io_win_unmap(&sw);
    chunk_free(inv, scn, sb->header.chunk_size);
  }
  sen_io_win_unmap(&dw);
  return sen_success;
}

sen_rc
sen_inv_updspec_add(sen_inv_updspec *u, int pos, int32_t weight, int_least8_t boundary)
{
  struct _sen_inv_pos *p;
  if (!(p = SEN_MALLOC(sizeof(struct _sen_inv_pos)))) {
    return sen_memory_exhausted;
  }
  u->boundary &= boundary;
  u->score += weight;
  p->pos = pos;
  p->next = NULL;
  if (u->tail) {
    u->tail->next = p;
  } else {
    u->pos = p;
  }
  u->tail = p;
  u->tf++;
  return sen_success;
}

/* inv */

sen_inv *
sen_inv_create(const char *path, int flags, uint32_t initial_n_segments)
{
  int i;
  sen_io *seg, *chunk;
  sen_inv *inv;
  char path2[PATH_MAX];
  struct sen_inv_header *header;
  if (strlen(path) + 6 >= PATH_MAX) { return NULL; }
  strcpy(path2, path);
  strcat(path2, ".c");
  seg = sen_io_create(path, SEN_INV_HEADER_SIZE, SEN_INV_SEGMENT_SIZE,
		      MAX_SEGMENT, sen_io_auto, MAX_SEGMENT);
  if (!seg) { return NULL; }
  chunk = sen_io_create(path2, SEN_INV_HEADER_SIZE, SEN_INV_SEGMENT_SIZE,
			MAX_SEGMENT, sen_io_auto, MAX_SEGMENT);
  if (!chunk) {
    sen_io_close(seg);
    return NULL;
  }
  header = sen_io_header(seg);
  memcpy(header->idstr, SEN_INV_IDSTR, 16);
  for (i = 0; i < MAX_SEGMENT; i++) { header->segments[i] = 0; }
  header->initial_n_segments =
    initial_n_segments ? initial_n_segments : SEN_INV_INITIAL_N_SEGMENTS;
  if (!(inv = SEN_MALLOC(sizeof(sen_inv)))) {
    sen_io_close(seg);
    sen_io_close(chunk);
    return NULL;
  }
  inv->seg = seg;
  inv->chunk = chunk;
  inv->header = header;
  for (i = 0; i < MAX_SEGMENT; i++) {
    inv->ainfo[i].segno = -1;
    inv->ainfo[i].addr = NULL;
    inv->binfo[i].segno = -1;
    inv->binfo[i].addr = NULL;
  }
  return inv;
}

sen_inv *
sen_inv_open(const char *path)
{
  int i;
  sen_io *seg, *chunk;
  sen_inv *inv;
  char path2[PATH_MAX];
  struct sen_inv_header *header;
  if (strlen(path) + 6 >= PATH_MAX) { return NULL; }
  strcpy(path2, path);
  strcat(path2, ".c");
  seg = sen_io_open(path, sen_io_auto, MAX_SEGMENT);
  if (!seg) { return NULL; }
  chunk = sen_io_open(path2, sen_io_auto, MAX_SEGMENT);
  if (!chunk) {
    sen_io_close(seg);
    return NULL;
  }
  header = sen_io_header(seg);
  /*
  if (memcmp(header->idstr, SEN_INV_IDSTR, 16)) {
    sen_io_close(seg);
    sen_io_close(chunk);
    return NULL;
  }
  */
  if (!(inv = SEN_MALLOC(sizeof(sen_inv)))) {
    sen_io_close(seg);
    sen_io_close(chunk);
    return NULL;
  }
  inv->seg = seg;
  inv->chunk = chunk;
  inv->header = header;
  for (i = 0; i < MAX_SEGMENT; i++) {
    inv->ainfo[i].segno = -1;
    inv->ainfo[i].addr = NULL;
    inv->binfo[i].segno = -1;
    inv->binfo[i].addr = NULL;
  }
  load_all_segments(inv);
  return inv;
}

sen_rc
sen_inv_close(sen_inv *inv)
{
  sen_rc rc;
  if ((rc = sen_io_close(inv->seg))) { return rc; }
  if ((rc = sen_io_close(inv->chunk))) { return rc; }
  SEN_FREE(inv);
  return rc;
}

sen_rc
sen_inv_update(sen_inv *inv, uint32_t key, sen_inv_updspec *u, int hint)
{
  sen_rc r = sen_success;
  buffer *b;
  buffer_rec *br, *bs;
  buffer_term *bt;
  uint32_t pos, size, *a = array_get(inv, key);
  // sen_log("key=%d b=%d tf=%d pos0=%d rid=%d", key, u->boundary, u->tf, u->pos->pos, u->rid);
  if (!u->tf) { return sen_invalid_argument; }
  if (!a) { return sen_memory_exhausted; }
  if (!(bs = encode_rec(u, &size))) { return sen_memory_exhausted; }
  if (*a) {
    if (*a & 1) {
      uint32_t size2, v = *a;
      buffer_rec *bs2;
      sen_inv_updspec u2;
      struct _sen_inv_pos pos2;
      pos2.pos = BIT10_01(v);
      pos2.next = NULL;
      u2.pos = &pos2;
      u2.rid = BIT29_11(v);
      u2.boundary = BIT31_30(v);
      u2.score = 1;
      u2.tf = 1;
      bs2 = encode_rec(&u2, &size2);
      if ((r = buffer_alloc(inv, size + size2, &pos, &bt, &br, &b, hint))) {
	SEN_FREE(bs2);
	goto exit;
      }
      bt->tid = key;
      // { sen_log("-bt=%p tid=%d", bt, key); }
      bt->pos_in_chunk = 0;
      bt->pos_in_buffer = (uintptr_t)br - (uintptr_t)b;
      memcpy(br, bs2, size2);
      br = (buffer_rec *)(((uintptr_t)br) + size2);
      SEN_FREE(bs2);
    } else {
      if ((r = buffer_at(inv, *a, &bt, &b))) { goto exit; }
      if (b->header.buffer_free < size) {
	sen_log("flushing b=%p free=%d, seg(%d)", b, b->header.buffer_free, *a >> W_OF_SEGMENT);
	buffer_flush(inv, *a >> W_OF_SEGMENT);
	if ((r = buffer_at(inv, *a, &bt, &b))) { goto exit; }
	sen_log("flushed! b=%p free=%d, seg(%d)", b, b->header.buffer_free, *a >> W_OF_SEGMENT);
	if (b->header.buffer_free < size) {
	  sen_log("buffer(%d) is full (%d < %d) in sen_inv_update",
		  *a, b->header.buffer_free, size);
	  /* todo: must be splitted */
	  r = sen_memory_exhausted;
	  goto exit;
	}
      }
      b->header.buffer_free -= size;
      br = ((void *)&b->terms[b->header.nterms]) + b->header.buffer_free;
    }
  } else {
    if (u->rid < 0x80000 && u->tf == 1 && u->score == 1 && u->pos->pos < 0x400) {
      *a = (u->boundary << 30) + (u->rid << 11) + (u->pos->pos << 1) + 1;
      goto exit;
    } else {
      if ((r = buffer_alloc(inv, size, &pos, &bt, &br, &b, hint))) { goto exit; }
      bt->tid = key;
      bt->pos_in_chunk = 0;
      bt->pos_in_buffer = 0;
    }
  }
  // sen_log("*a=%d,pos=%d", *a, pos);
  /*
  br->next = bt->pos_in_buffer;
  br->rid = u->rid;
  br->score = u->score;
  br->tf = u->tf;
  p = (uint32_t *)NEXT_ADDR(br);
  for (pp = u->pos; pp; pp = pp->next) {
    *p++ = pp->pos;
  }
  */
  memcpy(br, bs, size);
  br->var1 += bt->pos_in_buffer;
  bt->pos_in_buffer = (uintptr_t)br - (uintptr_t)b;
  if (!*a || (*a & 1)) { *a = pos; }
exit :
  SEN_FREE(bs);
  return r;
}

sen_rc
sen_inv_delete(sen_inv *inv, uint32_t key, sen_inv_updspec *u)
{
  sen_rc r = sen_success;
  buffer *b;
  buffer_rec *br;
  buffer_term *bt;
  uint32_t *a = array_at(inv, key);
  if (!a || !*a) { goto exit; }
  if (*a & 1) {
    if (u->rid == BIT29_11(*a)) { *a = 0; }
    goto exit;
  } 
  if ((r = buffer_at(inv, *a, &bt, &b))) { goto exit; }
  if (b->header.buffer_free < sizeof(buffer_rec)) {
    sen_log("flushing b=%p free=%d, seg(%d)", b, b->header.buffer_free, *a >> W_OF_SEGMENT);
    buffer_flush(inv, *a >> W_OF_SEGMENT);
    if ((r = buffer_at(inv, *a, &bt, &b))) { goto exit; }
    sen_log("flushed! b=%p free=%d, seg(%d)", b, b->header.buffer_free, *a >> W_OF_SEGMENT);
    if (b->header.buffer_free < sizeof(buffer_rec)) {
      /* todo: must be splitted ? */
      sen_log("buffer(%d) is full (%d < %d) in sen_inv_delete",
	      *a, b->header.buffer_free, sizeof(buffer_rec));
      r = sen_memory_exhausted;
      goto exit;
    }
  }
  b->header.buffer_free -= sizeof(buffer_rec);
  br = ((void *)&b->terms[b->header.nterms]) + b->header.buffer_free;
  br->var0 = B31 + u->rid;
  br->var1 = bt->pos_in_buffer;
  bt->pos_in_buffer = (uintptr_t)br - (uintptr_t)b;

exit :
  return r;
}

sen_inv_entry *
sen_inv_fetch(sen_inv *inv, uint32_t key)
{
  sen_io_win iw;
  sen_inv_entry *r  = NULL;
  uint32_t *a = array_at(inv, key);
  if (a && *a) {
    if (*a & 1) {
      uint32_t v = *a;
      sen_log("df=1,, pos=%d, rid=%d", BIT10_01(v), BIT29_11(v));
    } else {
      uint8_t *sp = NULL, *spe = NULL, *p;
      uint32_t cn, next, rid, bdf, i;
      sen_set *h;
      struct _sen_inv_set_entry **hh, *hp = NULL;
      buffer *buf;
      buffer_rec *br, **brp;
      buffer_term *bt;

      uint32_t ib = 0, ndf = 0, srid = 0, lrid = 0, brid, tf, score_bound, gap;

      if (buffer_at(inv, *a, &bt, &buf)) { return r; }

      if ((cn = buf->header.chunk) != CHUNK_NOT_ASSIGNED) {
	sp = sen_io_win_map(inv->chunk, &iw,
			    cn, bt->pos_in_chunk, bt->size_in_chunk, sen_io_rdonly);
	if (!sp) { return r; }
      }

      h = sen_set_open(sizeof(int), sizeof(buffer_rec *), 0);
      for (next = bt->pos_in_buffer; next; next = BIT17_00(br->var1)) {
	br = (buffer_rec *)(((void *)buf) + next);
	rid = BIT26_00(br->var0);
	if (!sen_set_get(h, &rid, NULL)) {
	  sen_set_add(h, &rid, (void **)&brp);
	  *brp = br;
	}
      }
      if (sp) { spe = sp + bt->size_in_chunk; }
      bdf = h->n_entries;
      hh = (struct _sen_inv_set_entry **)sen_set_sort(h, 0, NULL);

      sen_log("bdf=%d cn=%d pos_in_chunk=%d size_in_chunk=%d", bdf, cn, bt->pos_in_chunk, bt->size_in_chunk);

#define PUTNEXTS2() { \
  if (srid) { \
    sen_log("srid=%d", srid);\
    gap = srid - lrid; \
    ndf++; \
    DECODE(tf, sp); \
    DECODE(score_bound, sp); \
    sen_log("tf=%d score=%d", tf, score_bound);\
    while (tf--) { DECODE(gap, sp);\
      sen_log("gap=%d", gap);\
    } \
    lrid = srid; \
  } \
  if (sp < spe) { \
    DECODE(gap, sp); \
    srid += gap; \
  } else { \
    srid = 0; \
  } \
  /* sen_log("gap=%d srid=%d", gap, srid); */ \
}
#define PUTNEXTB2() { \
  br = hp->br; \
  sen_log("brid=%d", brid); \
  gap = brid - lrid; \
  if (gap <= 0) { sen_log("hp->rid(%d) <= lrid(%d)", hp->rid, lrid); } \
  ndf++; \
  sen_log("gap=%d", gap); \
  if (BIT31(br->var0)) { \
    sen_log("tf=%d, score=%d, pos=%d", 1, (br->var0 >> 27), BIT31_18(br->var1)); \
  } else if (BIT31(br->var1)) { \
    sen_log("tf=%d, score=%d, pos1=%d ", 2, BIT30_27(br->var0), BIT30_18(br->var1)); \
    p = NEXT_ADDR(br); \
    DECODE(gap, p); \
    sen_log("pos2=%d", gap + (BIT30_18(br->var1))); \
  } else if (BIT30(br->var0)) { \
    tf = BIT29_27(br->var0); \
    sen_log("tf=%d, score=%d ", tf, BIT31_18(br->var1)); \
    p = NEXT_ADDR(br); \
    for (i = 0; i < tf; i++) { \
      DECODE(gap, p); \
      sen_log("gap=%d", gap); \
    } \
  } else { \
    p = NEXT_ADDR(br); \
    DECODE(tf, p); \
    DECODE(score_bound, p); \
    sen_log("tf=%d, score=%d ", tf, score_bound); \
    for (i = 0; i < tf; i++) { \
      DECODE(gap, p); \
      sen_log("gap=%d", gap); \
    } \
  } \
  lrid = brid; \
  if (ib < bdf) { \
    hp = hh[ib++]; \
    brid = hp->rid; \
  } else { \
    brid = 0; \
  } \
}

      GETNEXTS();
      GETNEXTB();
      for (;;) {
	if (brid) {
	  if (srid) {
	    if (brid > srid) {
	      PUTNEXTS2();
	    } else {
	      if (brid == srid) { GETNEXTS(); }
	      PUTNEXTB2();
	    }
	  } else {
	    PUTNEXTB2();
	  }
	} else {
	  if (srid) {
	    PUTNEXTS2();
	  } else {
	    break;
	  }
	}
      }

      if (hh) { SEN_FREE(hh); }
      sen_set_close(h);
      if (cn != CHUNK_NOT_ASSIGNED) { sen_io_win_unmap(&iw); }
    }
  }
  return r;
}

sen_inv_cursor *
sen_inv_cursor_open(sen_inv *inv, uint32_t key)
{
  sen_inv_cursor *c  = NULL;
  uint32_t *a = array_at(inv, key);
  if (!a || !*a) { return c; }
  if (!(c = SEN_MALLOC(sizeof(sen_inv_cursor)))) { return c; }
  memset(c, 0, sizeof(sen_inv_cursor));
  if (*a & 1) {
    uint32_t v = *a;
    c->yet = 1;
    c->tf = 1;
    c->score = 1;
    c->boundary = BIT31_30(v);
    c->rid = BIT29_11(v);
    c->pos = BIT10_01(v);
  } else {
    uint32_t next, rid, chunk;
    buffer *buf;
    buffer_rec *br, **brp;
    buffer_term *bt;
    if (buffer_at(inv, *a, &bt, &buf)) {
      SEN_FREE(c);
      return NULL;
    }
    if (bt->size_in_chunk && (chunk = buf->header.chunk) != CHUNK_NOT_ASSIGNED) {
      c->sp = sen_io_win_map(inv->chunk, &c->iw,
			     chunk, bt->pos_in_chunk, bt->size_in_chunk, sen_io_rdonly);
      if (!c->sp) {
	SEN_FREE(c);
	return NULL;
      }
      c->spe = c->sp + bt->size_in_chunk;
    }
    c->h = sen_set_open(sizeof(int), sizeof(buffer_rec *), 0);
    for (next = bt->pos_in_buffer; next; next = BIT17_00(br->var1)) {
      br = (buffer_rec *)(((void *)buf) + next);
      rid = BIT26_00(br->var0);
      if (!sen_set_get(c->h, &rid, NULL)) {
	sen_set_add(c->h, &rid, (void **)&brp);
	*brp = br;
      }
    }
    if (c->h->n_entries) {
      c->hp = c->hh = (struct _sen_inv_set_entry **)sen_set_sort(c->h, 0, NULL);
      c->hpe = c->hh + c->h->n_entries;
    }
  }
  return c;
}

static inline sen_rc
_sen_inv_cursor_next(sen_inv_cursor *c)
{
  sen_rc rc = sen_success;
  if (!c->h) { /* (*a & 1) */
    if (c->yet) {
      c->yet = 0;
    } else {
      rc = sen_other_error;
    }
  } else {
    uint32_t srid = c->srid;
    uint32_t brid = c->brid;
    if (c->rid == srid) {
      uint8_t *sp = c->sp;
      if (c->yet) {
	uint32_t tf;
	DECODE(tf, sp);
	SKIP(sp);
	while (tf--) { SKIP(sp); }
      }
      if (sp < c->spe) {
	uint32_t gap;
	DECODE(gap, sp);
	srid += gap;
	c->yet = 1;
      } else {
	srid = 0;
      }
      c->sp = sp;
    }
    if (c->rid == brid) {
      if (brid) { c->hp++; }
      if (c->hp < c->hpe) {
	brid = (*(c->hp))->rid;
      } else {
	brid = 0;
      }
    }
    if (brid) {
      if (srid) {
	if (brid > srid) {
	  c->rid = srid;
	} else {
	  c->rid = brid;
	}
      } else {
	c->rid = brid;
      }
    } else {
      if (srid) {
	c->rid = srid;
      } else {
	rc = sen_other_error;
      }
    }
    c->srid = srid;
    c->brid = brid;
  }
  return rc;
}

sen_rc
sen_inv_cursor_next(sen_inv_cursor *c)
{
  sen_rc rc;
  buffer_rec *br;
  while ((rc = _sen_inv_cursor_next(c)) == sen_success) {
    if (!c->h) { break; } 
    if (c->rid != c->brid) { break; }
    br = (*(c->hp))->br;
    if (BIT31_29(br->var0) != 0x04) { break; }
  }
  return rc;
}

sen_rc
sen_inv_cursor_get_property(sen_inv_cursor *c)
{
  uint32_t i;
  if (!c->h) { /* (*a & 1) */ return sen_success; }
  if (c->rid == c->brid) {
    buffer_rec *br = (*(c->hp))->br;
    if (BIT31(br->var0)) { // tf=0~1, score<4, pos<16384
      c->score = BIT30_29(br->var0);
      c->boundary = BIT28_27(br->var0);
      c->tf = c->score ? 1 : 0;
    } else if (BIT31(br->var1)) { // tf=2, score<4, pos0<8192
      c->tf = 2;
      c->score = BIT30_29(br->var0);
      c->boundary = BIT28_27(br->var0);
    } else if (BIT30(br->var0)) { // tf<8, score<2048
      c->tf = BIT29_27(br->var0);
      c->score = BIT30_20(br->var1);
      c->boundary = BIT19_18(br->var1);
    } else {
      uint8_t *p = NEXT_ADDR(br);
      DECODE(c->tf, p);
      DECODE(c->score, p);
      c->boundary = c->score & 0x03;
      c->score = c->score >> 2;
    }
  } else {
    uint8_t *sp = c->sp;
    if (!c->yet) { return sen_invalid_argument; }
    DECODE(c->tf, sp);
    DECODE(c->score, sp);
    c->boundary = c->score & 0x03;
    c->score = c->score >> 2;
    for (i = 0; i < c->tf; i++) { SKIP(sp); }
    c->sp = sp;
    c->yet = 0;
  }
  return sen_success;
}

static inline int
sen_inv_postings_set(sen_inv_postings *p, int pos, int count)
{
  if (pos < 0 || p->size <= pos) { 
    return 0;
  }
  if (pos + 1 > p->mark) { p->mark = pos + 1; }
  return ++(p->array[pos]) == count;
}

int
sen_inv_cursor_get_postings(sen_inv_cursor *c, sen_inv_postings *p,
			    int offset, int count)
{
  // todo: check and expand size of p->array...
  int rc = 0;
  uint32_t i, gap, pos;
  if (!c->h) { /* (*a & 1) */
    pos = c->pos - offset;
    if (sen_inv_postings_set(p, pos, count)) { rc++; }
  } else {
    if (c->rid == c->brid) {
      uint8_t *bp;
      buffer_rec *br = (*(c->hp))->br;
      if (BIT31(br->var0)) { // tf=0~1, score<4, pos<16384
	c->score = BIT30_29(br->var0);
	c->boundary = BIT28_27(br->var0);
	if (c->score) {
	  c->tf = 1;
	  pos = BIT31_18(br->var1) - offset;
	  if (sen_inv_postings_set(p, pos, count)) { rc++; }
	} else {
	  c->tf = 0;
	}
      } else if (BIT31(br->var1)) { // tf=2, score<4, pos0<8192
	c->tf = 2;
	c->score = BIT30_29(br->var0);
	c->boundary = BIT28_27(br->var0);
	pos = BIT30_18(br->var1) - offset;
	if (sen_inv_postings_set(p, pos, count)) { rc++; }
	bp = NEXT_ADDR(br);
	DECODE(gap, bp);
	pos += gap;
	if (sen_inv_postings_set(p, pos, count)) { rc++; }
      } else if (BIT30(br->var0)) { // tf<8, score<2048
	c->tf = BIT29_27(br->var0);
	c->score = BIT30_20(br->var1);
	c->boundary = BIT19_18(br->var1);
	pos = -offset;
	bp = NEXT_ADDR(br);
	if (c->tf > 10000) { sen_log("huge tf! (%d)", c->tf); }
	for (i = 0; i < c->tf; i++) {
	  DECODE(gap, bp);
	  pos += gap;
	  if (sen_inv_postings_set(p, pos, count)) { rc++; }
	}
      } else {
	bp = NEXT_ADDR(br);
	DECODE(c->tf, bp);
	DECODE(c->score, bp);
	c->boundary = c->score & 0x03;
	c->score = c->score >> 2;
	if (c->tf > 10000) { sen_log("huge tf!(%d), rid=%d", c->tf, c->rid); }
	for (i = 0, pos = -offset; i < c->tf; i++) {
	  DECODE(gap, bp);
	  pos += gap;
	  if (sen_inv_postings_set(p, pos, count)) { rc++; }
	}
      }
    } else {
      uint8_t *sp = c->sp;
      if (!c->yet) { return rc; }
      DECODE(c->tf, sp);
      DECODE(c->score, sp);
      c->boundary = c->score & 0x03;
      c->score = c->score >> 2;
      if (c->tf > 10000) { sen_log("huge tf %d", c->tf); }
      for (i = 0, pos = -offset; i < c->tf; i++) {
	DECODE(gap, sp);
	pos += gap;
	if (sen_inv_postings_set(p, pos, count)) { rc++; }
      }
      c->sp = sp;
      c->yet = 0;
    }
  }
  return rc;
}

sen_rc
sen_inv_cursor_close(sen_inv_cursor *c)
{
  if (!c) { return sen_invalid_argument; }
  if (c->hh) { SEN_FREE(c->hh); }
  if (c->sp) { sen_io_win_unmap(&c->iw); }
  if (c->h) { sen_set_close(c->h); }
  SEN_FREE(c);
  return sen_success;
}

void
sen_inv_describe_term(sen_inv *inv, uint32_t key)
{
  uint32_t *a = array_at(inv, key);
  if (a && *a) {
    if (*a & 1) {
      sen_log("%8d %8d %8d", 1, 0, 0);
    } else {
      buffer *buf;
      buffer_term *bt;
      if (buffer_at(inv, *a, &bt, &buf)) { return; }
      sen_log("%8d %8d %8d", bt->size_in_chunk, buf->header.chunk_size, bt->pos_in_buffer);
      {
	uint32_t score, tf, boundary, next, rid, pos;
	buffer_rec *br;
	uint8_t *bp;
	for (next = bt->pos_in_buffer; next; next = BIT17_00(br->var1)) {
	  br = (buffer_rec *)(((void *)buf) + next);
	  rid = BIT26_00(br->var0);
	  if (BIT31(br->var0)) { // tf=0~1, score<4, pos<16384
	    score = BIT30_29(br->var0);
	    boundary = BIT28_27(br->var0);
	    if (score) {
	      tf = 1;
	      pos = BIT31_18(br->var1);
	    } else {
	      tf = 0;
	      pos = 0;
	    }
	  } else if (BIT31(br->var1)) { // tf=2, score<4, pos0<8192
	    tf = 2;
	    score = BIT30_29(br->var0);
	    boundary = BIT28_27(br->var0);
	    pos = BIT30_18(br->var1);
	  } else if (BIT30(br->var0)) { // tf<8, score<2048
	    tf = BIT29_27(br->var0);
	    score = BIT30_20(br->var1);
	    boundary = BIT19_18(br->var1);
	    bp = NEXT_ADDR(br);
	    DECODE(pos, bp);
	  } else {
	    bp = NEXT_ADDR(br);
	    DECODE(tf, bp);
	    DECODE(score, bp);
	    boundary = score & 0x03;
	    score = score >> 2;
	    DECODE(pos, bp);
	  }
	  sen_log("%d %d %d %d", tf, score, boundary, pos);
	}
      }
    }
  } else {
    sen_log("%8d %8d %8d", 0, 0, 0);
  }
}

uint32_t
sen_inv_estimate_size(sen_inv *inv, uint32_t key)
{
  uint32_t *a = array_at(inv, key);
  if (a && *a) {
    if (*a & 1) {
      return 1;
    } else {
      buffer *buf;
      buffer_term *bt;
      if (buffer_at(inv, *a, &bt, &buf)) { return 0; }
      return bt->size_in_chunk + (bt->pos_in_buffer ? 2 : 0);
    }
  } else {
    return 0;
  }
}

#ifndef NO_DEBUG
inline static sen_rc
decode_rec(buffer_rec *br)
{
  uint8_t *p;
  uint32_t rid, tf, boundary, score, pos, gap, i, next;
  rid = BIT26_00(br->var0);
  next = BIT17_00(br->var1);
  sen_log("next=%d, rid=%d ", next, rid);
  if (BIT31(br->var0)) { // tf=0~1, score<4, pos<16384
    score = BIT30_29(br->var0);
    boundary = BIT28_27(br->var0);
    if (score) {
      tf = 1;
      pos = BIT31_18(br->var1);
    } else {
      tf = 0;
      pos = 0;
    }
    sen_log("score=%d, tf=%d, pos=%d", score, tf, pos);
  } else if (BIT31(br->var1)) { // tf=2, score<4, pos0<8192
    tf = 2;
    score = BIT30_29(br->var0);
    boundary = BIT28_27(br->var0);
    pos = BIT30_18(br->var1);
    sen_log("score=%d, tf=%d, pos=%d : ", score, tf, pos);
    p = NEXT_ADDR(br);
    DECODE(gap, p);
    pos += gap;
    sen_log("pos2=%d", pos);
  } else if (BIT30(br->var0)) { // tf<8, score<2048
    tf = BIT29_27(br->var0);
    score = BIT30_20(br->var1);
    boundary = BIT19_18(br->var1);
    sen_log("score=%d, tf=%d : (", score, tf);
    p = NEXT_ADDR(br);
    for (i = 0, pos = 0; i < tf; i++) {
      DECODE(gap, p);
      pos += gap;
      sen_log("%d ", pos);
    }
    sen_log(")");
  } else {
    p = NEXT_ADDR(br);
    DECODE(tf, p);
    DECODE(score, p);
    boundary = score & 0x03;
    score = score >> 2;
    sen_log("score=%d, tf=%d :: (", score, tf);
    for (i = 0, pos = 0; i < tf; i++) {
      DECODE(gap, p);
      pos += gap;
      sen_log("%d ", pos);
    }
    sen_log(")");
  }
  return sen_success;
}
#endif /* NO_DEBUG */
