/* Copyright(C) 2004 Brazil

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA */

#include <sys/types.h>
#include <sys/mman.h>
#include <string.h>
#include <limits.h>
#include "set.h"

#define INDEX_SIZE (256 * sizeof(void *))
#define INITIAL_N_ENTRIES 256
#define INITIAL_INDEX_SIZE 256

#if defined(MAP_ANON) && !defined(MAP_ANONYMOUS)
#define MAP_ANONYMOUS MAP_ANON
#endif

#define GARBAGE ((entry *) 1)

typedef struct _sen_set_entry entry;

struct _sen_set_entry {
  uint32_t key;
  union {
    struct {
      char * key;
      uintptr_t val;
    } str;
    uint8_t bin[1];
  } u;
};

inline static entry *
entry_new(sen_set *set)
{
  entry *e;
  if (set->garbages) {
    e = set->garbages;
    set->garbages = (entry *) e->key;
    e->key = 0;
  } else {
    void *chunk = set->chunks[set->curr_chunk];
    if (!chunk) {
      chunk = SEN_MMAP(set->entry_size * (INITIAL_N_ENTRIES << set->curr_chunk),
		   PROT_READ|PROT_WRITE, MAP_PRIVATE|MAP_ANONYMOUS, -1, 0);
      if (chunk == MAP_FAILED) { return NULL; }
      set->chunks[set->curr_chunk] = chunk;
    }
    e = chunk + set->entry_size * set->curr_entry;
    if (++set->curr_entry >= (INITIAL_N_ENTRIES << set->curr_chunk)) {
      set->curr_chunk++;
      set->curr_entry = 0;
    }
  }
  return e;
}

sen_set *
sen_set_open(uint32_t key_size, uint32_t value_size, uint32_t index_size)
{
  sen_set *set;
  uint32_t entry_size, i;
  for (i = INITIAL_INDEX_SIZE; i < index_size; i <<= 1);
  switch (key_size) {
  case 0 :
    entry_size = sizeof(uint32_t) + sizeof(uintptr_t) + value_size;
    break;
  case sizeof(uint32_t) :
    entry_size = sizeof(uint32_t) + value_size;
    break;
  default :
    entry_size = sizeof(uint32_t) + key_size + value_size;
    break;
  }
  while (entry_size % sizeof(uintptr_t)) { entry_size++; }
  if (!(set = SEN_MALLOC(sizeof(sen_set)))) { return NULL; }
  memset(set, 0, sizeof(sen_set));
  set->key_size = key_size;
  set->value_size = value_size;
  set->entry_size = entry_size;
  set->max_offset = i - 1;
  set->index = SEN_MMAP(i * sizeof(entry *),
			 PROT_READ|PROT_WRITE, MAP_PRIVATE|MAP_ANONYMOUS, -1, 0);
  if (set->index == MAP_FAILED) {
    SEN_FREE(set);
    return NULL;
  }
  return set;
}

static sen_rc
sen_set_reset(sen_set * set)
{
  entry **index;
  uint32_t m, n;
  {
    uint32_t ne = set->n_entries;
    if (ne > INT_MAX) { return sen_memory_exhausted; }
    for (n = INITIAL_INDEX_SIZE, ne *= 2; n <= ne; n <<= 1);
  }
  index = SEN_MMAP(n * sizeof(entry *),
		   PROT_READ|PROT_WRITE, MAP_PRIVATE|MAP_ANONYMOUS, -1, 0);
  if (index == MAP_FAILED) { return sen_memory_exhausted; }
  m = n - 1;
  {
    uint32_t i, j;
    entry *e, **sp, **dp;
    for (j = set->max_offset + 1, sp = set->index; j; j--, sp++) {
      e = *sp;
      if (!e || (e == GARBAGE)) { continue; }
      for (i = e->key; ; i += 1048573) {
	dp = index + (i & m);
	if (!*dp) { break; }
      }
      *dp = e;
    }
  }
  {
    entry **i0 = set->index;
    uint32_t n0 = set->max_offset + 1;
    set->index = index;
    set->max_offset = m;
    set->n_garbages = 0;
    SEN_MUNMAP(i0, n0 * sizeof(entry *));
  }
  return sen_success;
}

sen_rc
sen_set_close(sen_set * set)
{
  uint32_t i;
  if (!set->key_size) {
    entry *e, **sp;
    for (i = set->max_offset + 1, sp = set->index; i; i--, sp++) {
      e = *sp;
      if (!e || (e == GARBAGE)) { continue; }
      if (e->u.str.key) {
	SEN_FREE(e->u.str.key);
      }
    }
  }
  for (i = 0; i < SEN_SET_MAX_CHUNK; i++) {
    if (set->chunks[i]) {
      SEN_MUNMAP(set->chunks[i], set->entry_size * (INITIAL_N_ENTRIES << i));
    }
  }
  SEN_MUNMAP(set->index, (set->max_offset + 1) * sizeof(entry *));
  SEN_FREE(set);
  return sen_success;
}

inline static uint32_t
str_set(const unsigned char *p)
{
  uint32_t r;
  for (r = 0; *p; p++) { r = (r * 1021) + *p; }
  return r;
}

inline static uint32_t
bin_set(const uint8_t *p, uint32_t length)
{
  uint32_t r;
  for (r = 0; length--; p++) { r = (r * 1021) + *p; }
  return r;
}

sen_set_he *
sen_set_int_at(sen_set *set, const uint32_t *key, void **value)
{
  entry *e, **ep, **index = set->index;
  uint32_t h = *key, i, m = set->max_offset;
  for (i = h; ep = index + (i & m), (e = *ep); i += 1048573) {
    if (e == GARBAGE) { continue; }
    if (e->key == h) {
      if (value) { *value = &e->u; }
      return ep;
    }
  }
  return NULL;
}

sen_set_he *
sen_set_str_at(sen_set *set, const char *key, void **value)
{
  entry *e, **ep, **index = set->index;
  uint32_t h = str_set(key), i, m = set->max_offset;
  for (i = h; ep = index + (i & m), (e = *ep); i += 1048573) {
    if (e == GARBAGE) { continue; }
    if (e->key == h && !strcmp(key, e->u.str.key)) {
      if (value) { *value = &e->u.str.val; }
      return ep;
    }
  }
  return NULL;
}

sen_set_he *
sen_set_bin_at(sen_set *set, const void *key, void **value)
{
  entry *e, **ep, **index = set->index;
  uint32_t h = bin_set(key, set->key_size), i, m = set->max_offset;
  for (i = h; ep = index + (i & m), (e = *ep); i += 1048573) {
    if (e == GARBAGE) { continue; }
    if (e->key == h && !memcmp(key, e->u.bin, set->key_size)) {
      if (value) { *value = &e->u.bin[set->key_size]; }
      return ep;
    }
  }
  return NULL;
}

sen_set_he *
sen_set_get(sen_set *set, const void *key, void **value)
{
  switch (set->key_size) {
  case 0 :
    return sen_set_str_at(set, key, value);
  case sizeof(uint32_t) :
    return sen_set_int_at(set, key, value);
  default :
    return sen_set_bin_at(set, key, value);
  }
}

sen_set_he *
sen_set_int_get(sen_set *set, const uint32_t *key, void **value)
{
  entry *e, **ep, **np = NULL, **index = set->index;
  uint32_t h = *key, i, m = set->max_offset;
  for (i = h; ep = index + (i & m), e = *ep; i += 1048573) {
    if (e == GARBAGE) {
      if (!np) { np = ep; }
    } else {
      if (e->key == h) { goto exit; }
    }
  }
  if (np) {
    set->n_garbages--;
    ep = np;
  }
  e = entry_new(set);
  e->key = h;
  *ep = e;
  set->n_entries++;
exit :
  if (value) { *value = &e->u; }
  return ep;
}

sen_set_he *
sen_set_str_get(sen_set *set, const char *key, void **value)
{
  entry *e, **ep, **np = NULL, **index = set->index;
  uint32_t h = str_set(key), i, m = set->max_offset;
  for (i = h; ep = index + (i & m), e = *ep; i += 1048573) {
    if (e == GARBAGE) {
      if (!np) { np = ep; }
    } else {
      if (e->key == h && !strcmp(key, e->u.str.key)) { goto exit; }
    }
  }
  if (np) {
    set->n_garbages--;
    ep = np;
  }
  e = entry_new(set);
  e->key = h;
  e->u.str.key = strdup(key);
  *ep = e;
  set->n_entries++;
exit :
  if (value) { *value = &e->u.str.val; }
  return ep;
}

sen_set_he *
sen_set_bin_get(sen_set *set, const void *key, void **value)
{
  entry *e, **ep, **np = NULL, **index = set->index;
  uint32_t h = bin_set(key, set->key_size), i, m = set->max_offset;
  for (i = h; ep = index + (i & m), e = *ep; i += 1048573) {
    if (e == GARBAGE) {
      if (!np) { np = ep; }
    } else {
      if (e->key == h && !memcmp(key, e->u.bin, set->key_size)) {
	goto exit;
      }
    }
  }
  if (np) {
    set->n_garbages--;
    ep = np;
  }
  e = entry_new(set);
  e->key = h;
  memcpy(e->u.bin, key, set->key_size);
  *ep = e;
  set->n_entries++;
exit :
  if (value) { *value = &e->u.bin[set->key_size]; }
  return ep;
}

sen_set_he *
sen_set_add(sen_set *set, const void *key, void **value)
{
  if ((set->n_entries + set->n_garbages) * 2 > set->max_offset) {
    sen_set_reset(set);
  }
  switch (set->key_size) {
  case 0 :
    return sen_set_str_get(set, key, value);
  case sizeof(uint32_t) :
    return sen_set_int_get(set, key, value);
  default :
    return sen_set_bin_get(set, key, value);
  }
}

sen_rc
sen_set_delete(sen_set *set, sen_set_he *ep)
{
  entry *e;
  if (!ep || !*ep) { return sen_invalid_argument; }
  e = *ep;
  *ep = GARBAGE;
  if (!set->key_size) { SEN_FREE(e->u.str.key); }
  memset(e, 0, set->entry_size);
  e->key = (uint32_t) set->garbages;
  set->garbages = e;
  set->n_entries--;
  set->n_garbages++;
  return sen_success;
}

sen_set_cursor *
sen_set_cursor_open(sen_set *set)
{
  sen_set_cursor *c;
  if (!(c = SEN_MALLOC(sizeof(sen_set_cursor)))) { return NULL; }
  c->set = set;
  c->index = set->index;
  c->curr = set->index;
  c->rest = set->max_offset + 1;
  return c;
}

sen_set_he *
sen_set_cursor_next(sen_set_cursor *c, void **key, void **value)
{
  uint32_t i;
  entry *e, **ep;
  if (!c->rest || c->index != c->set->index) { return NULL; }
  for (i = c->rest, ep = c->curr;;i--, ep++) {
    if (!i) {
      c->rest = 0;
      return NULL;
    }
    e = *ep;
    if (e && e != GARBAGE) { break; }
  }
  switch (c->set->key_size) {
  case 0 :
    if (key) { *key = e->u.str.key; }
    if (value) { *value = &e->u.str.val; }
    break;
  case sizeof(uint32_t) :
    if (key) { *key = &e->key; }
    if (value) { *value = &e->u; }
    break;
  default :
    if (key) { *key = e->u.bin; }
    if (value) { *value = &e->u.bin[c->set->key_size]; }
    break;
  }
  c->curr = ep + 1;
  c->rest = i - 1;
  return ep;
}

sen_rc
sen_set_cursor_close(sen_set_cursor *cursor)
{
  SEN_FREE(cursor);
  return sen_success;
}

static inline void
swap(entry **a, entry **b)
{
  entry *c = *a;
  *a = *b;
  *b = c;
}

inline static entry **
pack_int(sen_set *set, entry **res)
{
  uint32_t i, n, ck, m = set->max_offset;
  entry **head, **tail, *e, *c;
  for (i = m >> 1;; i = (i + 1) & m) {
    if ((c = set->index[i]) && (c != GARBAGE)) { break; }
  }
  head = res;
  n = set->n_entries - 1;
  tail = res + n;
  ck = c->key;
  while (n--) {
    for (;;) {
      i = (i + 1) & m;
      if ((e = set->index[i]) && (e != GARBAGE)) { break; }
    }
    if (e->key < ck) {
      *head++ = e;
    } else {
      *tail-- = e;
    }
  }
  *head = c;
  return set->n_entries > 2 ? head : NULL;
}

inline static entry **
part_int(entry **b, entry **e)
{
  uint32_t ck;
  int d = e - b;
  entry **c;
  if ((*b)->key > (*e)->key) { swap(b, e); }
  if (d < 2) { return NULL; }
  c = b + (d >> 1);
  if ((*b)->key > (*c)->key) {
    swap(b, c);
  } else {
    if ((*c)->key > (*e)->key) { swap(c, e); }
  }
  if (d < 3) { return NULL; }
  b++;
  ck = (*c)->key;
  swap(b, c);
  c = b;
  for (;;) {
    while ((*++b)->key < ck) ;
    while ((*--e)->key > ck) ;
    if (b >= e) { break; }
    swap(b, e);
  }
  swap(c, e);
  return e;
}

static void
_sort_int(entry **head, entry **tail, int range)
{
  int rest;
  entry **c;
  if (head < tail && (c = part_int(head, tail))) {
    rest = range - 1 - (c - head);
    _sort_int(head, c - 1, range);
    if (rest > 0) { _sort_int(c + 1, tail, rest); }
  }
}

inline static void
sort_int(sen_set *set, entry **res, int range)
{
  entry **c = pack_int(set, res);
  if (c) {
    int rest = range - 1 - (c - res);
    _sort_int(res, c - 1, range);
    if (rest > 0 ) { _sort_int(c + 1, res + set->n_entries - 1, rest); }
  }
}

sen_set_he *
sen_set_sort(sen_set *set, int range,
	     int(*compar)(const sen_set_he *, const sen_set_he *))
{
  entry **res;
  if (!set->n_entries) { return NULL; }
  if (!(res = SEN_MALLOC(sizeof(entry *) * set->n_entries))) {
    return NULL;
  }
  if (!range) { range = set->n_entries; }
  switch (set->key_size) {
  case 0 :
    sort_int(set, res, range);
    break; // todo
  case sizeof(uint32_t) :
    sort_int(set, res, range);
    break;
  default :
    break; // todo
  }
  return res;
}
