/* Copyright(C) 2004 Brazil

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA */

#ifndef _SENNA_H
#define _SENNA_H

#ifdef	__cplusplus
extern "C" {
#endif

#define SEN_INDEX_NORMALIZE     0x0001
#define SEN_INDEX_SPLIT_ALPHA   0x0002
#define SEN_INDEX_SPLIT_DIGIT   0x0004
#define SEN_INDEX_SPLIT_SYMBOL  0x0008
#define SEN_INDEX_NGRAM         0x0010

#define SEN_STACK_SIZE       0x10000000

typedef enum {
  sen_success = 0,
  sen_memory_exhausted,
  sen_invalid_format,
  sen_file_operation_error,
  sen_invalid_argument,
  sen_other_error
} sen_rc;

typedef enum {
  sen_enc_default = 0,
  sen_enc_none,
  sen_enc_euc_jp,
  sen_enc_utf8,
  sen_enc_sjis,
} sen_encoding;

typedef struct _sen_set sen_set;
typedef struct _sen_sym sen_sym;
typedef struct _sen_inv sen_inv;
typedef struct _sen_index sen_index;
typedef struct _sen_records sen_records;
typedef struct _sen_set_cursor sen_set_cursor;
typedef struct _sen_set_entry * sen_set_he;

/******** basic API ********/

sen_rc sen_init(void);

/* sen_index class */

struct _sen_index {
  int foreign_keyp;
  sen_sym *keys;
  sen_sym *lexicon;
  sen_inv *inv;
};

sen_index *sen_index_create(const char *path, int key_size,
			    int flags, int initial_n_segments, sen_encoding encoding);
sen_index *sen_index_open(const char *path);
sen_rc sen_index_close(sen_index *i);
sen_rc sen_index_remove(const char *path);
sen_rc sen_index_upd(sen_index *i, const void *key,
                     const char *oldvalue, const char *newvalue);
sen_records *sen_index_sel(sen_index *i, const char *string);

/* sen_records class */

struct _sen_records {
  sen_index *index;
  sen_set *records;
  sen_set_cursor *cursor;
  const void *curr_key;
  unsigned curr_score;
};

const void * sen_records_next(sen_records *r);
sen_rc sen_records_rewind(sen_records *r);
int sen_records_curr_score(sen_records *r);
const void * sen_records_curr_key(sen_records *r);
int sen_records_nhits(sen_records *r);
int sen_records_find(sen_records *r, const void *key);
sen_rc sen_records_close(sen_records *r);

/******** advanced API ********/

typedef struct _sen_values sen_values;
typedef struct _sen_query sen_query;

/* sen_query class */

sen_query *sen_query_open(sen_index *i);
sen_rc sen_query_close(sen_query *q);
sen_rc sen_query_add(sen_query *q, const char *string,
                     int weight, char operation, int recall);
sen_records *sen_query_exec(sen_query *q);

/* sen_values class */

sen_values *sen_values_open(sen_index *i);
sen_rc sen_value_close(sen_values *v);
sen_rc sen_values_add(sen_values *v, const char *val, unsigned int weight);
sen_rc sen_index_update(sen_index *i, const void *key, unsigned int section,
                        sen_values *oldvalues, sen_values *newvalues);

/******** low level API ********/

/* sen_set class */

sen_set *sen_set_open(unsigned key_size, unsigned value_size, unsigned index_size);
sen_rc sen_set_close(sen_set *set);
sen_set_he *sen_set_add(sen_set *set, const void *key, void **value);
sen_set_he *sen_set_get(sen_set *set, const void *key, void **value);
sen_rc sen_set_del(sen_set *set, sen_set_he *e);
sen_set_cursor *sen_set_cursor_open(sen_set *set);
sen_set_he *sen_set_cursor_next(sen_set_cursor *cursor, void **key, void **value);
sen_rc sen_set_cursor_close(sen_set_cursor *cursor);
sen_set_he *sen_set_sort(sen_set *set, int range,
                         int(*compar)(const sen_set_he *, const sen_set_he *));
sen_set *sen_set_union(sen_set *a, sen_set *b);
sen_set *sen_set_subtract(sen_set *a, sen_set *b);
sen_set *sen_set_intersect(sen_set *a, sen_set *b);
unsigned int sen_set_intersect2(sen_set *a, sen_set *b);

/* sen_sym class */

typedef unsigned sen_id;

sen_sym * sen_sym_create(const char *path, unsigned key_size,
			 unsigned flags, sen_encoding encoding);
sen_sym * sen_sym_open(const char *path);
sen_rc sen_sym_close(sen_sym *sym);
sen_rc sen_sym_remove(const char *path);
sen_id sen_sym_add(sen_sym *sym, const unsigned char *key);
sen_id sen_sym_get(sen_sym *sym, const unsigned char *key);
sen_rc sen_sym_del(sen_sym *sym, const unsigned char *key);
unsigned sen_sym_size(sen_sym *sym);
const unsigned char *sen_sym_key(sen_sym *sym, sen_id id);
sen_set * sen_sym_prefix_search(sen_sym *sym, const unsigned char *key);
sen_set * sen_sym_suffix_search(sen_sym *sym, const unsigned char *key);
sen_id sen_sym_common_prefix_search(sen_sym *sym, const unsigned char *key);

/* sen_index additional functions */

sen_index *sen_index_create_with_keys(const char *path, sen_sym *keys,
			    int flags, int initial_n_segments, sen_encoding encoding);
sen_index *sen_index_open_with_keys(const char *path, sen_sym *keys);

/* misc */

#define SEN_MALLOC malloc
#define SEN_REALLOC realloc
#define SEN_FREE free
#define SEN_MMAP(length,prot,flags,fd,offset) (mmap(0,length,prot,flags,fd,offset))
#define SEN_MUNMAP(start,length) (munmap(start,length))

void sen_log(char *fmt, ...);

#ifdef __cplusplus
}
#endif

#endif /* _SENNA_H */
